# vim: set fileencoding=utf8 :
from __future__ import with_statement
from __future__ import unicode_literals
import os
import random
import string
import codecs
from six import StringIO, BytesIO
import logging
import socket

from pytds.tds_types import TimeType, DateTime2Type, DateType, DateTimeOffsetType, BitType, TinyIntType, SmallIntType, \
    IntType, BigIntType, RealType, FloatType, NVarCharType, VarBinaryType, SmallDateTimeType, DateTimeType, DecimalType, \
    MoneyType, UniqueIdentifierType, VariantType, ImageType, VarBinaryMaxType, VarCharType, TextType, NTextType, \
    NVarCharMaxType, VarCharMaxType, XmlType

try:
    import unittest2 as unittest
except:
    import unittest
import sys
from decimal import Decimal, getcontext
import logging
from time import sleep
from datetime import datetime, date, time
import uuid
import pytest
import pytds.tz
import pytds.login
import pytds.smp
tzoffset = pytds.tz.FixedOffsetTimezone
utc = pytds.tz.utc
import pytds.extensions
import six
from six.moves import xrange
from pytds import (
    connect, ProgrammingError, TimeoutError, Time,
    Error, IntegrityError, Timestamp, DataError, Date, Binary,
    output, default,
    STRING, BINARY, NUMBER, DATETIME, DECIMAL, INTEGER, REAL, XML)
from pytds.tds_types import (DateTimeSerializer, SmallMoneyType)
from pytds.tds_base import (
    Column,
    IS_TDS73_PLUS, IS_TDS71_PLUS,
    )
import dbapi20
import pytds
import settings


logger = logging.getLogger(__name__)

LIVE_TEST = getattr(settings, 'LIVE_TEST', True)


def create_test_database():
    if not LIVE_TEST:
        return
    logger.info('in setup class')
    kwargs = settings.CONNECT_KWARGS.copy()
    kwargs['database'] = 'master'
    kwargs['autocommit'] = True
    with connect(**kwargs) as conn:
        with conn.cursor() as cur:
            try:
                cur.execute('drop database [{0}]'.format(settings.DATABASE))
            except:
                logger.exception('Failed to drop database')
                pass
            try:
                cur.execute('create database [{0}]'.format(settings.DATABASE))
            except:
                pass
            try:
                cur.execute('create schema myschema')
            except:
                pass
            try:
                cur.execute('create table myschema.bulk_insert_table(num int, data varchar(100))')
            except:
                pass
            try:
                cur.execute('''
                create procedure testproc (@param int, @add int = 2, @outparam int output)
                as
                begin
                    set nocount on
                    --select @param
                    set @outparam = @param + @add
                    return @outparam
                end
                ''')
            except:
                pass


create_test_database()


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
def test_connection_timeout_with_mars():
    kwargs = settings.CONNECT_KWARGS.copy()
    kwargs['database'] = 'master'
    kwargs['timeout'] = 1
    kwargs['use_mars'] = True
    with connect(*settings.CONNECT_ARGS, **kwargs) as conn:
        cur = conn.cursor()
        with pytest.raises(TimeoutError):
            cur.execute("waitfor delay '00:00:05'")
        cur.execute('select 1')


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
def test_connection_no_mars_autocommit():
    kwargs = settings.CONNECT_KWARGS.copy()
    kwargs.update({
        'use_mars': False,
        'timeout': 1,
        'pooling': True,
        'autocommit': True,
    })
    with connect(**kwargs) as conn:
        with conn.cursor() as cur:
            # test execute scalar with empty response
            cur.execute_scalar('declare @tbl table(f int); select * from @tbl')

            cur.execute("print 'hello'")
            messages = cur.messages
            assert len(messages) == 1
            assert len(messages[0]) == 2
            # in following assert exception class does not have to be exactly as specified
            assert messages[0][0] == pytds.OperationalError
            assert messages[0][1].text == 'hello'
            assert messages[0][1].line == 1
            assert messages[0][1].severity == 0
            assert messages[0][1].number == 0
            assert messages[0][1].state == 1
            assert 'hello' in messages[0][1].message

        # test cursor usage after close, should raise exception
        cur = conn.cursor()
        cur.execute_scalar('select 1')
        cur.close()
        with pytest.raises(Error) as ex:
            cur.execute('select 1')
        assert 'Cursor is closed' in str(ex.value)
        # calling get_proc_return_status on closed cursor works
        # this test does not have to pass
        assert cur.get_proc_return_status() is None
        # calling rowcount on closed cursor works
        # this test does not have to pass
        assert cur.rowcount == -1
        # calling description on closed cursor works
        # this test does not have to pass
        assert cur.description is None
        # calling messages on closed cursor works
        # this test does not have to pass
        assert cur.messages is None
        # calling description on closed cursor works
        # this test does not have to pass
        assert cur.native_description is None


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
def test_connection_timeout_no_mars():
    kwargs = settings.CONNECT_KWARGS.copy()
    kwargs.update({
        'use_mars': False,
        'timeout': 1,
        'pooling': True,
    })
    with connect(**kwargs) as conn:
        with conn.cursor() as cur:
            with pytest.raises(TimeoutError):
                cur.execute("waitfor delay '00:00:05'")
        with conn.cursor() as cur:
            cur.execute("select 1")
            cur.fetchall()

        # test cancelling
        with conn.cursor() as cur:
            cur.execute('select 1')
            cur.cancel()
            assert cur.fetchall() == []
            cur.execute('select 2')
            assert cur.fetchall() == [(2,)]

        # test rollback
        conn.rollback()

        # test callproc on non-mars connection
        with conn.cursor() as cur:
            cur.callproc('sp_reset_connection')

        with conn.cursor() as cur:
            # test spid property on non-mars cursor
            assert cur.spid is not None

            # test tzinfo_factory property r/w
            cur.tzinfo_factory = cur.tzinfo_factory

    # test non-mars cursor with connection pool enabled
    with connect(**kwargs) as conn:
        with conn.cursor() as cur:
            cur.execute('select 1')
            assert cur.fetchall() == [(1,)]


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
def test_connection_no_mars_no_pooling():
    kwargs = settings.CONNECT_KWARGS.copy()
    kwargs.update({
        'use_mars': False,
        'pooling': False,
    })
    with connect(**kwargs) as conn:
        with conn.cursor() as cur:
            cur.execute("select 1")
            assert cur.fetchall() == [(1,)]


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
def test_row_strategies():
    kwargs = settings.CONNECT_KWARGS.copy()
    kwargs.update({
        'row_strategy': pytds.list_row_strategy,
    })
    with connect(**kwargs) as conn:
        with conn.cursor() as cur:
            cur.execute("select 1")
            assert cur.fetchall() == [[1]]
    kwargs.update({
        'row_strategy': pytds.namedtuple_row_strategy,
    })
    import collections
    with connect(**kwargs) as conn:
        with conn.cursor() as cur:
            cur.execute("select 1 as f")
            assert cur.fetchall() == [collections.namedtuple('Row', ['f'])(1)]
    kwargs.update({
        'row_strategy': pytds.recordtype_row_strategy,
    })
    with connect(**kwargs) as conn:
        with conn.cursor() as cur:
            cur.execute("select 1 as e, 2 as f")
            row, = cur.fetchall()
            assert row.e == 1
            assert row.f == 2
            assert row[0] == 1
            assert row[:] == (1, 2)
            row[0] = 3
            assert row[:] == (3, 2)


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
def test_get_instances():
    if not hasattr(settings, 'BROWSER_ADDRESS'):
        return unittest.skip('BROWSER_ADDRESS setting is not defined')
    pytds.tds.tds7_get_instances(settings.BROWSER_ADDRESS)


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class ConnectionTestCase(unittest.TestCase):
    def setUp(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['database'] = 'master'
        self.conn = connect(*settings.CONNECT_ARGS, **kwargs)

    def tearDown(self):
        self.conn.close()


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class NoMarsTestCase(unittest.TestCase):
    def setUp(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['database'] = 'master'
        kwargs['use_mars'] = False
        self.conn = connect(*settings.CONNECT_ARGS, **kwargs)

    def tearDown(self):
        self.conn.close()

    def test_commit(self):
        cursor = self.conn.cursor()
        cursor.execute('select 1')
        cursor.fetchall()
        self.conn.commit()


class TestCaseWithCursor(ConnectionTestCase):
    def setUp(self):
        super(TestCaseWithCursor, self).setUp()
        self.cursor = self.conn.cursor()

    #def test_mars_sessions_recycle_ids(self):
    #    if not self.conn.mars_enabled:
    #        self.skipTest('Only relevant to mars')
    #    for _ in xrange(2 ** 16 + 1):
    #        cur = self.conn.cursor()
    #        cur.close()

    def test_parameters_ll(self):
        _params_tests(self)



class TestVariant(ConnectionTestCase):
    def _t(self, result, sql):
        with self.conn.cursor() as cur:
            cur.execute("select cast({0} as sql_variant)".format(sql))
            val, = cur.fetchone()
            self.assertEqual(result, val)

    def test_new_datetime(self):
        if not IS_TDS73_PLUS(self.conn):
            self.skipTest('Requires TDS7.3+')
        import pytds.tz
        self._t(datetime(2011, 2, 3, 10, 11, 12, 3000), "cast('2011-02-03T10:11:12.003000' as datetime2)")
        self._t(time(10, 11, 12, 3000), "cast('10:11:12.003000' as time)")
        self._t(date(2011, 2, 3), "cast('2011-02-03' as date)")
        self._t(datetime(2011, 2, 3, 10, 11, 12, 3000, pytds.tz.FixedOffsetTimezone(3 * 60)), "cast('2011-02-03T10:11:12.003000+03:00' as datetimeoffset)")


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class BadConnection(unittest.TestCase):
    def test_invalid_parameters(self):
        with self.assertRaises(Error):
            with connect(server=settings.HOST + 'bad', database='master', user='baduser', password=settings.PASSWORD, login_timeout=1) as conn:
                with conn.cursor() as cur:
                    cur.execute('select 1')
        with self.assertRaises(Error):
            with connect(server=settings.HOST, database='doesnotexist', user=settings.USER, password=settings.PASSWORD) as conn:
                with conn.cursor() as cur:
                    cur.execute('select 1')
        with self.assertRaises(Error):
            with connect(server=settings.HOST, database='master', user='baduser', password=None) as conn:
                with conn.cursor() as cur:
                    cur.execute('select 1')

    def test_instance_and_port(self):
        host = settings.HOST
        if '\\' in host:
            host, _ = host.split('\\')
        with self.assertRaisesRegex(ValueError, 'Both instance and port shouldn\'t be specified'):
            with connect(server=host + '\\badinstancename', database='master', user=settings.USER, password=settings.PASSWORD, port=1212) as conn:
                with conn.cursor() as cur:
                    cur.execute('select 1')


def get_spid(conn):
    with conn.cursor() as cur:
        return cur.spid


def kill(conn, spid):
    with conn.cursor() as cur:
        cur.execute('kill {0}'.format(spid))


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class ConnectionClosing(unittest.TestCase):
    def test_open_close(self):
        for x in xrange(3):
            kwargs = settings.CONNECT_KWARGS.copy()
            kwargs['database'] = 'master'
            connect(**kwargs).close()

    def test_closing_after_closed_by_server(self):
        """
        You should be able to call close on connection closed by server
        """
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['database'] = 'master'
        kwargs['autocommit'] = True
        with connect(**kwargs) as master_conn:
            kwargs['autocommit'] = False
            with connect(**kwargs) as conn:
                with conn.cursor() as cur:
                    cur.execute('select 1')
                    conn.commit()
                    kill(master_conn, get_spid(conn))
                    sleep(0.2)
                conn.close()

    def test_connection_closed_by_server(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['database'] = 'master'
        with connect(**kwargs) as master_conn:
            master_conn.autocommit = True
            with connect(**kwargs) as conn:
                conn.autocommit = False
                # test overall recovery
                with conn.cursor() as cur:
                    cur.execute('select 1')
                    conn.commit()
                    kill(master_conn, get_spid(conn))
                    sleep(0.2)
                    cur.execute('select 1')
                    cur.fetchall()
                kill(master_conn, get_spid(conn))
                sleep(0.2)
                with conn.cursor() as cur:
                    cur.execute('select 1')

                # test cursor opening in a transaction, it should raise exception
                # make transaction dirty
                with conn.cursor() as cur:
                    cur.execute('select 1')
                kill(master_conn, get_spid(conn))
                sleep(0.2)
                # it does not have to raise this specific exception
                with pytest.raises(socket.error):
                    with conn.cursor() as cur:
                        cur.execute('select 1')

                # test recovery on transaction
                with conn.cursor() as cur:
                    cur.execute('create table ##testtable3 (fld int)')
                    kill(master_conn, get_spid(conn))
                    sleep(0.2)
                    with self.assertRaises(Exception):
                        cur.execute('select * from ##testtable2')
                        cur.fetchall()
                    conn.rollback()
                    cur.execute('select 1')

                # test server closed connection on rollback
                with conn.cursor() as cur:
                    cur.execute('select 1')
                kill(master_conn, get_spid(conn))
                sleep(0.2)
                conn.rollback()

            #with connect(server=settings.HOST, database='master', user=settings.USER, password=settings.PASSWORD) as conn:
            #    spid = get_spid(conn)
            #    with conn.cursor() as cur:
            #        # test recovery of specific lowlevel methods
            #        tds_submit_query(cur._session, "waitfor delay '00:00:05'; select 1")
            #        kill(master_conn, spid)
            #        self.assertTrue(cur._session.is_connected())
            #        with self.assertRaises(Exception):
            #            tds_process_tokens(cur._session, TDS_TOKEN_RESULTS)
            #        self.assertFalse(cur._session.is_connected())


#class EncryptionTest(unittest.TestCase):
#    def runTest(self):
#        conn = connect(server=settings.HOST, database='master', user=settings.USER, password=settings.PASSWORD, encryption_level=TDS_ENCRYPTION_REQUIRE)
#        cur = conn.cursor()
#        cur.execute('select 1')


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class SmallDateTimeTest(ConnectionTestCase):
    def _testval(self, val):
        with self.conn.cursor() as cur:
            cur.execute('select cast(%s as smalldatetime)', (val,))
            self.assertEqual(cur.fetchall(), [(val,)])

    def runTest(self):
        self._testval(Timestamp(2010, 2, 1, 10, 12, 0))
        self._testval(Timestamp(1900, 1, 1, 0, 0, 0))
        self._testval(Timestamp(2079, 6, 6, 23, 59, 0))
        with self.assertRaises(Error):
            self._testval(Timestamp(1899, 1, 1, 0, 0, 0))
        with self.assertRaises(Error):
            self._testval(Timestamp(2080, 1, 1, 0, 0, 0))


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class DateTimeTest(ConnectionTestCase):
    def _testencdec(self, val):
        self.assertEqual(val, DateTimeSerializer.decode(*DateTimeSerializer._struct.unpack(DateTimeSerializer.encode(val))))

    def _testval(self, val):
        with self.conn.cursor() as cur:
            cur.execute('select cast(%s as datetime)', (val,))
            self.assertEqual(cur.fetchall(), [(val,)])

    def runTest(self):
        with self.conn.cursor() as cur:
            cur.execute("select cast('9999-12-31T23:59:59.997' as datetime)")
            self.assertEqual(cur.fetchall(), [(Timestamp(9999, 12, 31, 23, 59, 59, 997000),)])
        self._testencdec(Timestamp(2010, 1, 2, 10, 11, 12))
        self._testval(Timestamp(2010, 1, 2, 0, 0, 0))
        self._testval(Timestamp(2010, 1, 2, 10, 11, 12))
        self._testval(Timestamp(1753, 1, 1, 0, 0, 0))
        self._testval(Timestamp(9999, 12, 31, 0, 0, 0))
        with self.conn.cursor() as cur:
            cur.execute("select cast(null as datetime)")
            self.assertEqual(cur.fetchall(), [(None,)])
        self._testval(Timestamp(9999, 12, 31, 23, 59, 59, 997000))
        with self.assertRaises(Error):
            self._testval(Timestamp(1752, 1, 1, 0, 0, 0))
        with self.conn.cursor() as cur:
            cur.execute('''
            if object_id('testtable') is not null
                drop table testtable
            ''')
            cur.execute('create table testtable (col datetime not null)')
            dt = Timestamp(2010, 1, 2, 20, 21, 22, 123000)
            cur.execute('insert into testtable values (%s)', (dt,))
            cur.execute('select col from testtable')
            self.assertEqual(cur.fetchone(), (dt,))


class NewDateTimeTest(ConnectionTestCase):
    def test_datetimeoffset(self):
        if not IS_TDS73_PLUS(self.conn):
            self.skipTest('Requires TDS7.3+')

        def _testval(val):
            with self.conn.cursor() as cur:
                import pytds.tz
                cur.tzinfo_factory = pytds.tz.FixedOffsetTimezone
                cur.execute('select cast(%s as datetimeoffset)', (val,))
                self.assertEqual(cur.fetchall(), [(val,)])

        with self.conn.cursor() as cur:
            import pytds.tz
            cur.tzinfo_factory = pytds.tz.FixedOffsetTimezone
            cur.execute("select cast('2010-01-02T20:21:22.1234567+05:00' as datetimeoffset)")
            self.assertEqual(datetime(2010, 1, 2, 20, 21, 22, 123456, tzoffset(5 * 60)), cur.fetchone()[0])
        _testval(Timestamp(2010, 1, 2, 0, 0, 0, 0, utc))
        _testval(Timestamp(2010, 1, 2, 0, 0, 0, 0, tzoffset(5 * 60)))
        _testval(Timestamp(1, 1, 1, 0, 0, 0, 0, utc))
        _testval(Timestamp(9999, 12, 31, 23, 59, 59, 999999, utc))
        _testval(Timestamp(2010, 1, 2, 0, 0, 0, 0, tzoffset(14)))
        _testval(Timestamp(2010, 1, 2, 0, 0, 0, 0, tzoffset(-14)))
        _testval(Timestamp(2010, 1, 2, 0, 0, 0, 0, tzoffset(-15)))

    def test_time(self):
        if not IS_TDS73_PLUS(self.conn):
            self.skipTest('Requires TDS7.3+')

        def testval(val):
            with self.conn.cursor() as cur:
                cur.execute('select cast(%s as time)', (val,))
                self.assertEqual(cur.fetchall(), [(val,)])

        testval(Time(14, 16, 18, 123456))
        testval(Time(0, 0, 0, 0))
        testval(Time(0, 0, 0, 0))
        testval(Time(0, 0, 0, 0))
        testval(Time(23, 59, 59, 999999))
        testval(Time(0, 0, 0, 0))
        testval(Time(0, 0, 0, 0))
        testval(Time(0, 0, 0, 0))

    def test_datetime2(self):
        if not IS_TDS73_PLUS(self.conn):
            self.skipTest('Requires TDS7.3+')

        def testval(val):
            with self.conn.cursor() as cur:
                cur.execute('select cast(%s as datetime2)', (val,))
                self.assertEqual(cur.fetchall(), [(val,)])

        testval(Timestamp(2010, 1, 2, 20, 21, 22, 345678))
        testval(Timestamp(2010, 1, 2, 0, 0, 0))
        testval(Timestamp(1, 1, 1, 0, 0, 0))
        testval(Timestamp(9999, 12, 31, 23, 59, 59, 999999))

    def test_date(self):
        if not IS_TDS73_PLUS(self.conn):
            self.skipTest('Requires TDS7.3+')

        def testval(val):
            with self.conn.cursor() as cur:
                cur.execute('select cast(%s as date)', (val,))
                self.assertEqual(cur.fetchall(), [(val,)])

        testval(Date(2010, 1, 2))
        testval(Date(2010, 1, 2))
        testval(Date(1, 1, 1))
        testval(Date(9999, 12, 31))


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class Auth(unittest.TestCase):
    @unittest.skipUnless(os.getenv('NTLM_USER') and os.getenv('NTLM_PASSWORD'), "requires NTLM_USER and NTLM_PASSWORD environment variables to be set")
    def test_ntlm(self):
        conn = connect(settings.HOST, auth=pytds.login.NtlmAuth(user_name=os.getenv('NTLM_USER'), password=os.getenv('NTLM_PASSWORD')))
        with conn.cursor() as cursor:
            cursor.execute('select 1')
            cursor.fetchall()

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_sspi(self):
        from pytds.login import SspiAuth
        with connect(settings.HOST, auth=SspiAuth()) as conn:
            with conn.cursor() as cursor:
                cursor.execute('select 1')
                cursor.fetchall()

    @unittest.skipIf(getattr(settings, 'SKIP_SQL_AUTH', False), 'SKIP_SQL_AUTH is set')
    def test_sqlauth(self):
        with connect(settings.HOST, user=settings.USER, password=settings.PASSWORD) as conn:
            with conn.cursor() as cursor:
                cursor.execute('select 1')
                cursor.fetchall()


class CloseCursorTwice(ConnectionTestCase):
    def runTest(self):
        cursor = self.conn.cursor()
        cursor.close()
        cursor.close()


class RegressionSuite(ConnectionTestCase):
    def test_cancel(self):
        self.conn.cursor().cancel()


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class TimezoneTests(unittest.TestCase):
    def check_val(self, conn, sql, input, output):
        with conn.cursor() as cur:
            cur.execute('select ' + sql, (input,))
            rows = cur.fetchall()
            self.assertEqual(rows[0][0], output)

    def runTest(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        use_tz = utc
        kwargs['use_tz'] = use_tz
        kwargs['database'] = 'master'
        with connect(*settings.CONNECT_ARGS, **kwargs) as conn:
            # Naive time should be interpreted as use_tz
            self.check_val(conn, '%s',
                           datetime(2011, 2, 3, 10, 11, 12, 3000),
                           datetime(2011, 2, 3, 10, 11, 12, 3000, utc))
            # Aware time shoule be passed as-is
            dt = datetime(2011, 2, 3, 10, 11, 12, 3000, tzoffset(1))
            self.check_val(conn, '%s', dt, dt)
            # Aware time should be converted to use_tz if not using datetimeoffset type
            dt = datetime(2011, 2, 3, 10, 11, 12, 3000, tzoffset(1))
            if IS_TDS73_PLUS(conn):
                self.check_val(conn, 'cast(%s as datetime2)', dt, dt.astimezone(use_tz))


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class DbapiTestSuite(dbapi20.DatabaseAPI20Test, ConnectionTestCase):
    driver = pytds
    connect_args = settings.CONNECT_ARGS
    connect_kw_args = settings.CONNECT_KWARGS

#    def _connect(self):
#        return connection

    def _try_run(self, *args):
        with self._connect() as con:
            with con.cursor() as cur:
                for arg in args:
                    cur.execute(arg)

    def _try_run2(self, cur, *args):
        for arg in args:
            cur.execute(arg)

    # This should create the "lower" sproc.
    def _callproc_setup(self, cur):
        self._try_run2(
            cur,
            """IF OBJECT_ID(N'[dbo].[to_lower]', N'P') IS NOT NULL DROP PROCEDURE [dbo].[to_lower]""",
            """
            CREATE PROCEDURE to_lower
                @input nvarchar(max)
            AS
            BEGIN
                select LOWER(@input)
            END
            """,
            )

    # This should create a sproc with a return value.
    def _retval_setup(self, cur):
        self._try_run2(
            cur,
            """IF OBJECT_ID(N'[dbo].[add_one]', N'P') IS NOT NULL DROP PROCEDURE [dbo].[add_one]""",
            """
CREATE PROCEDURE add_one (@input int)
AS
BEGIN
    return @input+1
END
""",
            )

    def test_retval(self):
        with self._connect() as con:
            cur = con.cursor()
            self._retval_setup(cur)
            values = cur.callproc('add_one', (1,))
            self.assertEqual(values[0], 1, 'input parameter should be left unchanged: %s' % (values[0],))

            self.assertEqual(cur.description, None, "No resultset was expected.")
            self.assertEqual(cur.return_value, 2, "Invalid return value: %s" % (cur.return_value,))

        # This should create a sproc with a return value.
    def _retval_select_setup(self, cur):
        self._try_run2(
            cur,
            """IF OBJECT_ID(N'[dbo].[add_one_select]', N'P') IS NOT NULL DROP PROCEDURE [dbo].[add_one_select]""",
            """
CREATE PROCEDURE add_one_select (@input int)
AS
BEGIN
    select 'a' as a
    select 'b' as b
    return @input+1
END
""",
            )

    def test_retval_select(self):
        with self._connect() as con:
            cur = con.cursor()
            self._retval_select_setup(cur)
            values = cur.callproc('add_one_select', (1,))
            self.assertEqual(values[0], 1, 'input parameter should be left unchanged: %s' % (values[0],))

            self.assertEqual(len(cur.description), 1, "Unexpected resultset.")
            self.assertEqual(cur.description[0][0], 'a', "Unexpected resultset.")
            self.assertEqual(cur.fetchall(), [('a',)], 'Unexpected resultset.')

            self.assertTrue(cur.nextset(), 'No second resultset found.')
            self.assertEqual(len(cur.description), 1, "Unexpected resultset.")
            self.assertEqual(cur.description[0][0], 'b', "Unexpected resultset.")

            self.assertEqual(cur.return_value, 2, "Invalid return value: %s" % (cur.return_value,))
            with self.assertRaises(Error):
                cur.fetchall()

    # This should create a sproc with an output parameter.
    def _outparam_setup(self, cur):
        self._try_run2(
            cur,
            """IF OBJECT_ID(N'[dbo].[add_one_out]', N'P') IS NOT NULL DROP PROCEDURE [dbo].[add_one_out]""",
            """
CREATE PROCEDURE add_one_out (@input int, @output int OUTPUT)
AS
BEGIN
    SET @output = @input+1
END
""",
            )

    def test_outparam(self):
        with self._connect() as con:
            cur = con.cursor()
            self._outparam_setup(cur)
            values = cur.callproc('add_one_out', (1, output(value=1)))
            self.assertEqual(len(values), 2, 'expected 2 parameters')
            self.assertEqual(values[0], 1, 'input parameter should be unchanged')
            self.assertEqual(values[1], 2, 'output parameter should get new values')

            values = cur.callproc('add_one_out', (None, output(value=1)))
            self.assertEqual(len(values), 2, 'expected 2 parameters')
            self.assertEqual(values[0], None, 'input parameter should be unchanged')
            self.assertEqual(values[1], None, 'output parameter should get new values')

    def test_assigning_select(self):
        # test that assigning select does not interfere with result sets
        with self._connect() as con:
            cur = con.cursor()
            cur.execute("""
declare @var1 int

select @var1 = 1
select @var1 = 2

select 'value'
""")
            self.assertFalse(cur.description)
            self.assertTrue(cur.nextset())

            self.assertFalse(cur.description)
            self.assertTrue(cur.nextset())

            self.assertTrue(cur.description)
            self.assertEqual([(u'value',)], cur.fetchall())
            self.assertFalse(cur.nextset())

            cur.execute("""
set nocount on

declare @var1 int

select @var1 = 1
select @var1 = 2

select 'value'
""")
            self.assertTrue(cur.description)
            self.assertEqual([(u'value',)], cur.fetchall())
            self.assertFalse(cur.nextset())

    # Don't need setoutputsize tests.
    def test_setoutputsize(self):
        pass

    def help_nextset_setUp(self, cur):
        self._try_run2(
            cur,
            """IF OBJECT_ID(N'[dbo].[deleteme]', N'P') IS NOT NULL DROP PROCEDURE [dbo].[deleteme]""",
            """
create procedure deleteme
as
begin
    select count(*) from %sbooze
    select name from %sbooze
end
""" % (self.table_prefix, self.table_prefix),
            )

    def help_nextset_tearDown(self, cur):
        cur.execute("drop procedure deleteme")

    def test_ExceptionsAsConnectionAttributes(self):
        pass

    def test_select_decimal_zero(self):
        with self._connect() as con:
            expected = (
                Decimal('0.00'),
                Decimal('0.0'),
                Decimal('-0.00'))

            cur = con.cursor()
            cur.execute("SELECT %s as A, %s as B, %s as C", expected)

            result = cur.fetchall()
            self.assertEqual(result[0], expected)

    def test_type_objects(self):
        with self._connect() as con:
            cur = con.cursor()
            cur.execute("""
select cast(0 as varchar),
       cast(1 as binary),
       cast(2 as int),
       cast(3 as real),
       cast(4 as decimal),
       cast('2005' as datetime),
       cast('6' as xml)
""")
            self.assertTrue(cur.description)
            col_types = [col[1] for col in cur.description]
            self.assertEqual(col_types[0], STRING)
            self.assertEqual(col_types[1], BINARY)
            self.assertEqual(col_types[2], NUMBER)
            self.assertEqual(col_types[2], INTEGER)
            self.assertEqual(col_types[3], NUMBER)
            self.assertEqual(col_types[3], REAL)
            # self.assertEqual(col_types[4], NUMBER) ?
            self.assertEqual(col_types[4], DECIMAL)
            self.assertEqual(col_types[5], DATETIME)
            self.assertEqual(col_types[6], XML)


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class TestBug4(unittest.TestCase):
    def test_as_dict(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['database'] = 'master'
        with connect(*settings.CONNECT_ARGS, **kwargs) as conn:
            conn.as_dict = True
            with conn.cursor() as cur:
                cur.execute('select 1 as a, 2 as b')
                self.assertDictEqual({'a': 1, 'b': 2}, cur.fetchone())


def _params_tests(self):
    def test_val(typ, val):
        with self.conn.cursor() as cur:
            param = Column(type=typ, value=val)
            logger.info("Testing with %s", repr(param))
            cur.execute('select %s', [param])
            self.assertTupleEqual(cur.fetchone(), (val,))
            self.assertIs(cur.fetchone(), None)

    test_val(BitType(), True)
    test_val(BitType(), False)
    test_val(BitType(), None)
    test_val(TinyIntType(), 255)
    test_val(SmallIntType(), 2 ** 15 - 1)
    test_val(IntType(), 2 ** 31 - 1)
    test_val(BigIntType(), 2 ** 63 - 1)
    test_val(IntType(), None)
    test_val(RealType(), 0.25)
    test_val(FloatType(), 0.25)
    test_val(RealType(), None)
    test_val(SmallDateTimeType(), datetime(1900, 1, 1, 0, 0, 0))
    test_val(SmallDateTimeType(), datetime(2079, 6, 6, 23, 59, 0))
    test_val(DateTimeType(), datetime(1753, 1, 1, 0, 0, 0))
    test_val(DateTimeType(), datetime(9999, 12, 31, 23, 59, 59, 990000))
    test_val(DateTimeType(), None)
    if pytds.tds_base.IS_TDS73_PLUS(self.conn._conn):
        test_val(DateType(), date(1, 1, 1))
        test_val(DateType(), date(9999, 12, 31))
        test_val(DateType(), None)
        test_val(TimeType(precision=0), time(0, 0, 0))
        test_val(TimeType(precision=6), time(23, 59, 59, 999999))
        test_val(TimeType(precision=0), None)
        test_val(DateTime2Type(precision=0), datetime(1, 1, 1, 0, 0, 0))
        test_val(DateTime2Type(precision=6), datetime(9999, 12, 31, 23, 59, 59, 999999))
        test_val(DateTime2Type(precision=0), None)
        test_val(DateTimeOffsetType(precision=6), datetime(9999, 12, 31, 23, 59, 59, 999999, utc))
        test_val(DateTimeOffsetType(precision=6), datetime(9999, 12, 31, 23, 59, 59, 999999, tzoffset(14)))
        test_val(DateTimeOffsetType(precision=0), datetime(1, 1, 1, 0, 0, 0, tzinfo=tzoffset(-14)))
        #test_val(DateTimeOffsetType(precision=0), datetime(1, 1, 1, 0, 0, 0, tzinfo=tzoffset(14)))
        test_val(DateTimeOffsetType(precision=6), None)
    test_val(DecimalType(scale=6, precision=38), Decimal('123.456789'))
    test_val(DecimalType(scale=6, precision=38), None)
    test_val(SmallMoneyType(), Decimal('-214748.3648'))
    test_val(SmallMoneyType(), Decimal('214748.3647'))
    test_val(MoneyType(), Decimal('922337203685477.5807'))
    test_val(MoneyType(), Decimal('-922337203685477.5808'))
    test_val(MoneyType(), None)
    test_val(UniqueIdentifierType(), None)
    test_val(UniqueIdentifierType(), uuid.uuid4())
    if pytds.tds_base.IS_TDS71_PLUS(self.conn._conn):
        test_val(VariantType(), None)
        #test_val(self.conn._conn.type_factory.SqlVariant(10), 100)
    test_val(VarBinaryType(size=10), b'')
    test_val(VarBinaryType(size=10), b'testtest12')
    test_val(VarBinaryType(size=10), None)
    test_val(VarBinaryType(size=8000), b'x' * 8000)
    test_val(VarCharType(size=10), None)
    test_val(VarCharType(size=10), '')
    test_val(VarCharType(size=10), 'test')
    test_val(VarCharType(size=8000), 'x' * 8000)
    test_val(NVarCharType(size=10), u'')
    test_val(NVarCharType(size=10), u'testtest12')
    test_val(NVarCharType(size=10), None)
    test_val(NVarCharType(size=4000), u'x' * 4000)
    test_val(TextType(), None)
    test_val(TextType(), '')
    test_val(TextType(), 'hello')
    test_val(NTextType(), None)
    test_val(NTextType(), '')
    test_val(NTextType(), 'hello')
    test_val(ImageType(), None)
    test_val(ImageType(), b'')
    test_val(ImageType(), b'test')
    if pytds.tds_base.IS_TDS72_PLUS(self.conn._conn):
        test_val(VarBinaryMaxType(), None)
        test_val(VarBinaryMaxType(), b'')
        test_val(VarBinaryMaxType(), b'testtest12')
        test_val(VarBinaryMaxType(), b'x' * (10 ** 6))
        test_val(NVarCharMaxType(), None)
        test_val(NVarCharMaxType(), 'test')
        test_val(NVarCharMaxType(), 'x' * (10 ** 6))
        test_val(VarCharMaxType(), None)
        test_val(VarCharMaxType(), 'test')
        test_val(VarCharMaxType(), 'x' * (10 ** 6))
        test_val(XmlType(), '<root/>')


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class TestTds70(unittest.TestCase):
    def setUp(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['database'] = 'master'
        kwargs['tds_version'] = pytds.tds_base.TDS70
        self.conn = connect(*settings.CONNECT_ARGS, **kwargs)

    def test_parsing(self):
        _params_tests(self)


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class TestTds71(unittest.TestCase):
    def setUp(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['database'] = 'master'
        kwargs['tds_version'] = pytds.tds_base.TDS71
        self.conn = connect(*settings.CONNECT_ARGS, **kwargs)

    def test_parsing(self):
        _params_tests(self)

    def test_transaction(self):
        self.conn.rollback()
        self.conn.commit()

    def test_bulk(self):
        f = StringIO("42\tfoo\n74\tbar\n")
        with self.conn.cursor() as cur:
            cur.copy_to(f, 'bulk_insert_table', schema='myschema', columns=('num', 'data'))
            cur.execute('select num, data from myschema.bulk_insert_table')
            self.assertListEqual(cur.fetchall(), [(42, 'foo'), (74, 'bar')])

    def test_call_proc(self):
        with self.conn.cursor() as cur:
            val = 45
            values = cur.callproc('testproc', (val, default, output(value=1)))
            #self.assertEqual(cur.fetchall(), [(val,)])
            self.assertEqual(val + 2, values[2])
            self.assertEqual(val + 2, cur.get_proc_return_status())


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class TestTds72(unittest.TestCase):
    def setUp(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['database'] = 'master'
        kwargs['tds_version'] = pytds.tds_base.TDS72
        self.conn = connect(*settings.CONNECT_ARGS, **kwargs)

    def test_parsing(self):
        _params_tests(self)


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class TestTds73A(unittest.TestCase):
    def setUp(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['database'] = 'master'
        kwargs['tds_version'] = pytds.tds_base.TDS73A
        self.conn = connect(*settings.CONNECT_ARGS, **kwargs)

    def test_parsing(self):
        _params_tests(self)


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class TestTds73B(unittest.TestCase):
    def setUp(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['database'] = 'master'
        kwargs['tds_version'] = pytds.tds_base.TDS73B
        self.conn = connect(*settings.CONNECT_ARGS, **kwargs)

    def test_parsing(self):
        _params_tests(self)

@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
class TestRawBytes(unittest.TestCase):
    def setUp(self):
        kwargs = settings.CONNECT_KWARGS.copy()
        kwargs['bytes_to_unicode'] = False
        kwargs['database'] = 'master'
        self.conn = connect(*settings.CONNECT_ARGS, **kwargs)

    def test_fetch(self):
        cur = self.conn.cursor()

        self.assertIsInstance(cur.execute_scalar("select cast('abc' as nvarchar(max))"), six.text_type)
        self.assertIsInstance(cur.execute_scalar("select cast('abc' as varchar(max))"), six.binary_type)
        self.assertIsInstance(cur.execute_scalar("select cast('abc' as text)"), six.binary_type)

        self.assertIsInstance(cur.execute_scalar("select %s", ['abc']), six.text_type)
        self.assertIsInstance(cur.execute_scalar("select %s", [b'abc']), six.binary_type)

        rawBytes = six.b('\x01\x02\x03')
        self.assertEqual(rawBytes, cur.execute_scalar("select cast(0x010203 as varchar(max))"))
        self.assertEqual(rawBytes, cur.execute_scalar("select %s", [rawBytes]))

        utf8char = six.b('\xee\xb4\xba')
        self.assertEqual(utf8char, cur.execute_scalar("select %s", [utf8char]))


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
def test_invalid_block_size():
    kwargs = settings.CONNECT_KWARGS.copy()
    kwargs.update({
        'blocksize': 4000,
    })
    with connect(**kwargs) as conn:
        with conn.cursor() as cur:
            cur.execute_scalar("select '{}'".format('x' * 8000))


@unittest.skipUnless(LIVE_TEST, "requires HOST variable to be set")
def test_readonly_connection():
    kwargs = settings.CONNECT_KWARGS.copy()
    kwargs.update({
        'readonly': True,
    })
    with connect(**kwargs) as conn:
        with conn.cursor() as cur:
            cur.execute_scalar("select 1")



import os

CONNECT_ARGS = []
CONNECT_KWARGS = {}

LIVE_TEST = 'HOST' in os.environ
if LIVE_TEST:
    HOST = os.environ['HOST']
    DATABASE = os.environ.get('DATABASE', 'test')
    USER = os.environ.get('SQLUSER', 'sa')
    PASSWORD = os.environ.get('SQLPASSWORD', 'sa')
    USE_MARS = bool(os.environ.get('USE_MARS', True))
    SKIP_SQL_AUTH = bool(os.environ.get('SKIP_SQL_AUTH'))

    import pytds

    CONNECT_KWARGS = {
        'server': HOST,
        'database': DATABASE,
        'user': USER,
        'password': PASSWORD,
        'use_mars': USE_MARS,
        'bytes_to_unicode': True,
        'pooling': True,
        'timeout': 30,
    }

    if 'tds_version' in os.environ:
        CONNECT_KWARGS['tds_version'] = getattr(pytds, os.environ['tds_version'])

    if 'auth' in os.environ:
        import pytds.login

        CONNECT_KWARGS['auth'] = getattr(pytds.login, os.environ['auth'])()

    if 'bytes_to_unicode' in os.environ:
        CONNECT_KWARGS['bytes_to_unicode'] = bool(os.environ.get('bytes_to_unicode'))



"""DB-SIG compliant module for communicating with MS SQL servers"""
from collections import deque
import datetime
import errno
import keyword
import os
import re
import six
import socket
import uuid
import warnings
import weakref
import logging

from six.moves import xrange

from pytds.tds_types import NVarCharType
from . import lcid
import pytds.tz
from .tds import (
    _TdsSocket, tds7_get_instances,
    _create_exception_by_message,
    output, default
)
from . import tds_base
from .tds_base import (
    Error, LoginError, DatabaseError, ProgrammingError,
    IntegrityError, DataError, InternalError,
    InterfaceError, TimeoutError, OperationalError,
    NotSupportedError, Warning, ClosedConnectionError,
    Column,
    PreLoginEnc)

from .tds_types import (
    TableValuedParam, Binary
)

from .tds_base import (
    ROWID, DECIMAL, STRING, BINARY, NUMBER, DATETIME, INTEGER, REAL, XML
)

from . import tls
import pkg_resources

__author__ = 'Mikhail Denisenko <denisenkom@gmail.com>'
__version__ = pkg_resources.get_distribution('python-tds').version

logger = logging.getLogger(__name__)


def _ver_to_int(ver):
    res = ver.split('.')
    if len(res) < 2:
        logger.warning('Invalid version {}, it should have 2 parts at least separated by "."'.format(ver))
        return 0
    maj, minor, _ = ver.split('.')
    return (int(maj) << 24) + (int(minor) << 16)


intversion = _ver_to_int(__version__)

#: Compliant with DB SIG 2.0
apilevel = '2.0'

#: Module may be shared, but not connections
threadsafety = 1

#: This module uses extended python format codes
paramstyle = 'pyformat'


class _TdsLogin:
    pass


def tuple_row_strategy(column_names):
    """ Tuple row strategy, rows returned as tuples, default
    """
    return tuple


def list_row_strategy(column_names):
    """  List row strategy, rows returned as lists
    """
    return list


def dict_row_strategy(column_names):
    """ Dict row strategy, rows returned as dictionaries
    """
    # replace empty column names with indices
    column_names = [(name or idx) for idx, name in enumerate(column_names)]

    def row_factory(row):
        return dict(zip(column_names, row))

    return row_factory


def is_valid_identifier(name):
    return name and re.match("^[_A-Za-z][_a-zA-Z0-9]*$", name) and not keyword.iskeyword(name)


def namedtuple_row_strategy(column_names):
    """ Namedtuple row strategy, rows returned as named tuples

    Column names that are not valid Python identifiers will be replaced
    with col<number>_
    """
    import collections
    # replace empty column names with placeholders
    column_names = [name if is_valid_identifier(name) else 'col%s_' % idx for idx, name in enumerate(column_names)]
    row_class = collections.namedtuple('Row', column_names)

    def row_factory(row):
        return row_class(*row)

    return row_factory


def recordtype_row_strategy(column_names):
    """ Recordtype row strategy, rows returned as recordtypes

    Column names that are not valid Python identifiers will be replaced
    with col<number>_
    """
    try:
        from namedlist import namedlist as recordtype  # optional dependency
    except ImportError:
        from recordtype import recordtype  # optional dependency
    # replace empty column names with placeholders
    column_names = [name if is_valid_identifier(name) else 'col%s_' % idx for idx, name in enumerate(column_names)]
    recordtype_row_class = recordtype('Row', column_names)

    # custom extension class that supports indexing
    class Row(recordtype_row_class):
        def __getitem__(self, index):
            if isinstance(index, slice):
                return tuple(getattr(self, x) for x in self.__slots__[index])
            return getattr(self, self.__slots__[index])

        def __setitem__(self, index, value):
            setattr(self, self.__slots__[index], value)

    def row_factory(row):
        return Row(*row)

    return row_factory


class _ConnectionPool(object):
    def __init__(self, max_pool_size=100, min_pool_size=0):
        self._max_pool_size = max_pool_size
        self._pool = {}

    def add(self, key, conn):
        l = self._pool.setdefault(key, []).append(conn)

    def take(self, key):
        l = self._pool.get(key, [])
        if len(l) > 0:
            return l.pop()
        else:
            return None


_connection_pool = _ConnectionPool()


class Connection(object):
    """Connection object, this object should be created by calling :func:`connect`"""

    def __init__(self):
        self._closed = False
        self._conn = None
        self._isolation_level = 0
        self._autocommit = True
        self._row_strategy = tuple_row_strategy
        self._login = None
        self._use_tz = None
        self._tzinfo_factory = None
        self._key = None
        self._pooling = False

    @property
    def as_dict(self):
        """
        Instructs all cursors this connection creates to return results
        as a dictionary rather than a tuple.
        """
        return self._row_strategy == dict_row_strategy

    @as_dict.setter
    def as_dict(self, value):
        if value:
            self._row_strategy = dict_row_strategy
        else:
            self._row_strategy = tuple_row_strategy

    @property
    def autocommit_state(self):
        """
        An alias for `autocommit`, provided for compatibility with pymssql
        """
        return self._autocommit

    def set_autocommit(self, value):
        """ An alias for `autocommit`, provided for compatibility with ADO dbapi
        """
        self.autocommit = value

    @property
    def autocommit(self):
        """
        The current state of autocommit on the connection.
        """
        return self._autocommit

    @autocommit.setter
    def autocommit(self, value):
        if self._autocommit != value:
            if value:
                if self._conn.tds72_transaction:
                    self._main_cursor._rollback(cont=False)
            else:
                self._main_cursor._begin_tran(isolation_level=self._isolation_level)
            self._autocommit = value

    @property
    def isolation_level(self):
        """Isolation level for transactions,
        for possible values see :ref:`isolation-level-constants`

        .. seealso:: `SET TRANSACTION ISOLATION LEVEL`__ in MSSQL documentation

            .. __: http://msdn.microsoft.com/en-us/library/ms173763.aspx
        """
        return self._isolation_level

    @isolation_level.setter
    def isolation_level(self, level):
        self._isolation_level = level

    def _assert_open(self):
        if self._closed:
            raise Error('Connection closed')
        if not self._conn or not self._conn.is_connected():
            self._open()

    def _trancount(self):
        with self.cursor() as cur:
            cur.execute('select @@trancount')
            return cur.fetchone()[0]

    @property
    def tds_version(self):
        """
        Version of tds protocol that is being used by this connection
        """
        self._assert_open()
        return self._conn.tds_version

    @property
    def product_version(self):
        """
        Version of the MSSQL server
        """
        self._assert_open()
        return self._conn.product_version

    @property
    def mars_enabled(self):
        """ Whether MARS is enabled or not on connection
        """
        return self._conn.mars_enabled

    def _connect(self, host, port, instance, timeout, sock=None):
        login = self._login

        try:
            login.server_name = host
            login.instance_name = instance
            port = _resolve_instance_port(
                host,
                port,
                instance,
                timeout=timeout)
            if not sock:
                logger.info('Opening socket to %s:%d', host, port)
                sock = socket.create_connection((host, port), timeout)
        except Exception as e:
            raise LoginError("Cannot connect to server '{0}': {1}".format(host, e), e)

        sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)

        # default keep alive should be 30 seconds according to spec:
        # https://msdn.microsoft.com/en-us/library/dd341108.aspx
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 30)

        sock.settimeout(timeout)
        conn = _TdsSocket(self._use_tz)
        self._conn = conn
        try:
            route = conn.login(login, sock, self._tzinfo_factory)
            if route is not None:
                # rerouted to different server
                sock.close()
                self._connect(host=route['server'],
                              port=route['port'],
                              instance=instance,
                              timeout=timeout)
                return

            if conn.mars_enabled:
                cursor = _MarsCursor(
                    self,
                    conn.create_session(self._tzinfo_factory),
                    self._tzinfo_factory)
            else:
                cursor = Cursor(
                    self,
                    conn.main_session,
                    self._tzinfo_factory)

            self._active_cursor = self._main_cursor = cursor
            if not self._autocommit:
                cursor._session.begin_tran(isolation_level=self._isolation_level)
            sock.settimeout(login.query_timeout)
        except:
            sock.close()
            raise

    def _try_open(self, timeout, sock=None):
        if self._pooling:
            res = _connection_pool.take(self._key)
            if res is not None:
                self._conn, sess = res
                if self._conn.mars_enabled:
                    cursor = _MarsCursor(
                        self,
                        sess,
                        self._tzinfo_factory)
                else:
                    cursor = Cursor(
                        self,
                        sess,
                        self._tzinfo_factory)
                self._active_cursor = self._main_cursor = cursor
                cursor.callproc('sp_reset_connection')
                return

        login = self._login
        host, port, instance = login.servers[0]
        self._connect(host=host, port=port, instance=instance, timeout=timeout, sock=sock)

    def _open(self, sock=None):
        import time
        self._conn = None
        self._dirty = False
        login = self._login
        connect_timeout = login.connect_timeout

        # using retry algorithm specified in
        # http://msdn.microsoft.com/en-us/library/ms175484.aspx
        retry_time = 0.08 * connect_timeout
        retry_delay = 0.2
        last_error = None
        end_time = time.time() + connect_timeout
        while True:
            for _ in xrange(len(login.servers)):
                try:
                    self._try_open(timeout=retry_time, sock=sock)
                    return
                except OperationalError as e:
                    last_error = e
                    # if there are more than one message this means
                    # that the login was successful, like in the
                    # case when database is not accessible
                    # mssql returns 2 messages:
                    # 1) Cannot open database "<dbname>" requested by the login. The login failed.
                    # 2) Login failed for user '<username>'
                    # in this case we want to retry
                    if self._conn is not None and len(self._conn.main_session.messages) <= 1:
                        # for the following error messages we don't retry
                        # because if the password is incorrect and we
                        # retry multiple times this can cause account
                        # to be locked
                        if e.msg_no in (
                                18456,  # login failed
                                18486,  # account is locked
                                18487,  # password expired
                                18488,  # password should be changed
                                18452,  # login from untrusted domain
                        ):
                            raise

                if time.time() > end_time:
                    raise last_error
                login.servers.rotate(-1)

            time.sleep(retry_delay)
            retry_time += 0.08 * connect_timeout
            retry_delay = min(1, retry_delay * 2)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def commit(self):
        """
        Commit transaction which is currently in progress.
        """
        self._assert_open()
        if self._autocommit:
            return
        if not self._conn.tds72_transaction:
            return
        self._main_cursor._commit(cont=True, isolation_level=self._isolation_level)

    def cursor(self):
        """
        Return cursor object that can be used to make queries and fetch
        results from the database.
        """
        self._assert_open()
        if self.mars_enabled:
            in_tran = self._conn.tds72_transaction
            if in_tran and self._dirty:
                try:
                    return _MarsCursor(self,
                                       self._conn.create_session(self._tzinfo_factory),
                                       self._tzinfo_factory)
                except (socket.error, OSError) as e:
                    self._conn.close()
                    raise
            else:
                try:
                    return _MarsCursor(self,
                                       self._conn.create_session(self._tzinfo_factory),
                                       self._tzinfo_factory)
                except (socket.error, OSError) as e:
                    if e.errno not in (errno.EPIPE, errno.ECONNRESET):
                        raise
                    self._conn.close()
                except ClosedConnectionError:
                    pass
                self._assert_open()
                return _MarsCursor(self,
                                   self._conn.create_session(self._tzinfo_factory),
                                   self._tzinfo_factory)
        else:
            return Cursor(self,
                          self._conn.main_session,
                          self._tzinfo_factory)

    def rollback(self):
        """
        Roll back transaction which is currently in progress.
        """
        try:
            if self._autocommit:
                return

            if not self._conn or not self._conn.is_connected():
                return

            if not self._conn.tds72_transaction:
                return

            self._main_cursor._rollback(cont=True,
                                        isolation_level=self._isolation_level)
        except socket.error as e:
            if e.errno in (errno.ENETRESET, errno.ECONNRESET, errno.EPIPE):
                return
            self._conn.close()
            raise
        except ClosedConnectionError:
            pass

    def close(self):
        """ Close connection to an MS SQL Server.

        This function tries to close the connection and free all memory used.
        It can be called more than once in a row. No exception is raised in
        this case.
        """
        if self._conn:
            if self._pooling:
                _connection_pool.add(self._key, (self._conn, self._main_cursor._session))
            else:
                self._conn.close()
            self._active_cursor = None
            self._main_cursor = None
            self._conn = None
        self._closed = True

    def _try_activate_cursor(self, cursor):
        if cursor is not self._active_cursor:
            session = self._active_cursor._session
            if session.in_cancel:
                session.process_cancel()

            if session.state == tds_base.TDS_PENDING:
                raise InterfaceError('Results are still pending on connection')
            self._active_cursor = cursor


class Cursor(six.Iterator):
    """
    This class represents a database cursor, which is used to issue queries
    and fetch results from a database connection.
    """
    def __init__(self, conn, session, tzinfo_factory):
        self._conn = weakref.ref(conn)
        self.arraysize = 1
        self._session = session
        self._tzinfo_factory = tzinfo_factory

    def _assert_open(self):
        conn = self._conn
        if conn is not None:
            conn = conn()
        if not conn:
            raise InterfaceError('Cursor is closed')
        conn._assert_open()
        self._session = conn._conn._main_session
        return conn

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __iter__(self):
        """
        Return self to make cursors compatibile with Python iteration
        protocol.
        """
        return self

    def _setup_row_factory(self):
        self._row_factory = None
        conn = self._conn()
        if self._session.res_info:
            column_names = [col[0] for col in self._session.res_info.description]
            self._row_factory = conn._row_strategy(column_names)

    def _callproc(self, procname, parameters):
        self._ensure_transaction()
        results = list(parameters)
        parameters = self._session._convert_params(parameters)
        self._exec_with_retry(lambda: self._session.submit_rpc(procname, parameters, 0))
        self._session.process_rpc()
        for key, param in self._session.output_params.items():
            results[key] = param.value
        self._setup_row_factory()
        return results

    def get_proc_outputs(self):
        """
        If stored procedure has result sets and OUTPUT parameters use this method
        after you processed all result sets to get values of OUTPUT parameters.
        :return: A list of output parameter values.
        """

        self._session.complete_rpc()
        results = [None] * len(self._session.output_params.items())
        for key, param in self._session.output_params.items():
            results[key] = param.value
        return results

    def callproc(self, procname, parameters=()):
        """
        Call a stored procedure with the given name.

        :param procname: The name of the procedure to call
        :type procname: str
        :keyword parameters: The optional parameters for the procedure
        :type parameters: sequence

        Note: If stored procedure has OUTPUT parameters and result sets this
        method will not return values for OUTPUT parameters, you should
        call get_proc_outputs to get values for OUTPUT parameters.
        """
        conn = self._assert_open()
        conn._try_activate_cursor(self)
        return self._callproc(procname, parameters)

    @property
    def return_value(self):
        """  Alias to :func:`get_proc_return_status`
        """
        return self.get_proc_return_status()

    @property
    def connection(self):
        """ Provides link back to :class:`Connection` of this cursor
        """
        return self._conn()

    @property
    def spid(self):
        """ MSSQL Server's SPID (session id)
        """
        return self._session._spid

    def _get_tzinfo_factory(self):
        return self._tzinfo_factory

    def _set_tzinfo_factory(self, tzinfo_factory):
        self._tzinfo_factory = self._session.tzinfo_factory = tzinfo_factory

    tzinfo_factory = property(_get_tzinfo_factory, _set_tzinfo_factory)

    def get_proc_return_status(self):
        """ Last stored proc result
        """
        if self._session is None:
            return None
        if not self._session.has_status:
            self._session.find_return_status()
        return self._session.ret_status if self._session.has_status else None

    def cancel(self):
        """ Cancel current statement
        """
        conn = self._assert_open()
        conn._try_activate_cursor(self)
        self._session.cancel_if_pending()

    def close(self):
        """
        Closes the cursor. The cursor is unusable from this point.
        """
        conn = self._conn
        if conn is not None:
            conn = conn()
        if conn is not None:
            if self is conn._active_cursor:
                conn._active_cursor = conn._main_cursor
                self._session = None
            self._conn = None

    def _exec_with_retry(self, fun):
        conn = self._assert_open()
        in_tran = conn._conn.tds72_transaction
        if in_tran and conn._dirty:
            conn._dirty = True
            try:
                return fun()
            except socket.error as e:
                if e.errno not in (errno.ECONNRESET, errno.EPIPE):
                    raise
                conn._conn.close()
        else:
            conn._dirty = True
            try:
                return fun()
            except socket.error as e:
                if e.errno not in (errno.ECONNRESET, errno.EPIPE):
                    raise
                conn._conn.close()
            except ClosedConnectionError:
                pass
            # in case of connection reset try again
            conn = self._assert_open()
            return fun()

    def _ensure_transaction(self):
        conn = self._conn()
        if not conn._autocommit and not conn._conn.tds72_transaction:
            conn._main_cursor._begin_tran(isolation_level=conn._isolation_level)

    def _execute(self, operation, params):
        self._ensure_transaction()
        operation = six.text_type(operation)
        if params:
            named_params = {}
            if isinstance(params, (list, tuple)):
                names = []
                pid = 1
                for val in params:
                    if val is None:
                        names.append('NULL')
                    else:
                        name = '@P{0}'.format(pid)
                        names.append(name)
                        named_params[name] = val
                        pid += 1
                if len(names) == 1:
                    operation = operation % names[0]
                else:
                    operation = operation % tuple(names)
            elif isinstance(params, dict):
                # prepend names with @
                rename = {}
                for name, value in params.items():
                    if value is None:
                        rename[name] = 'NULL'
                    else:
                        mssql_name = '@{0}'.format(name)
                        rename[name] = mssql_name
                        named_params[mssql_name] = value
                operation = operation % rename
            if named_params:
                named_params = self._session._convert_params(named_params)
                param_definition = u','.join(
                    u'{0} {1}'.format(p.column_name, p.type.get_declaration())
                    for p in named_params)
                self._exec_with_retry(lambda: self._session.submit_rpc(
                    tds_base.SP_EXECUTESQL,
                    [self._session.make_param('', operation), self._session.make_param('', param_definition)] + named_params,
                    0))
            else:
                self._exec_with_retry(lambda: self._session.submit_plain_query(operation))
        else:
            self._exec_with_retry(lambda: self._session.submit_plain_query(operation))
        self._session.find_result_or_done()
        self._setup_row_factory()

    def execute(self, operation, params=()):
        """ Execute the query

        :param operation: SQL statement
        :type operation: str
        """
        conn = self._assert_open()
        conn._try_activate_cursor(self)
        self._execute(operation, params)
        # for compatibility with pyodbc
        return self

    def _begin_tran(self, isolation_level):
        conn = self._assert_open()
        conn._try_activate_cursor(self)
        self._session.begin_tran(isolation_level=isolation_level)

    def _commit(self, cont, isolation_level=0):
        conn = self._assert_open()
        conn._try_activate_cursor(self)
        self._session.commit(cont=cont, isolation_level=isolation_level)
        conn._dirty = False

    def _rollback(self, cont, isolation_level=0):
        conn = self._assert_open()
        conn._try_activate_cursor(self)
        self._session.rollback(cont=cont, isolation_level=isolation_level)
        conn._dirty = False

    def executemany(self, operation, params_seq):
        counts = []
        for params in params_seq:
            self.execute(operation, params)
            if self._session.rows_affected != -1:
                counts.append(self._session.rows_affected)
        if counts:
            self._session.rows_affected = sum(counts)

    def execute_scalar(self, query_string, params=None):
        """
        This method sends a query to the MS SQL Server to which this object
        instance is connected, then returns first column of first row from
        result. An exception is raised on failure. If there are pending

        results or rows prior to executing this command, they are silently
        discarded.

        This method accepts Python formatting. Please see execute_query()
        for details.

        This method is useful if you want just a single value, as in:

            ``conn.execute_scalar('SELECT COUNT(*) FROM employees')``

        This method works in the same way as ``iter(conn).next()[0]``.
        Remaining rows, if any, can still be iterated after calling this
        method.
        """
        self.execute(query_string, params)
        row = self.fetchone()
        if not row:
            return None
        return row[0]

    def nextset(self):
        """ Move to next recordset in batch statement, all rows of current recordset are
        discarded if present.

        :returns: true if successful or ``None`` when there are no more recordsets
        """
        res = self._session.next_set()
        self._setup_row_factory()
        return res

    @property
    def rowcount(self):
        """ Number of rows affected by previous statement

        :returns: -1 if this information was not supplied by MSSQL server
        """
        if self._session is None:
            return -1
        return self._session.rows_affected

    @property
    def description(self):
        """ Cursor description, see http://legacy.python.org/dev/peps/pep-0249/#description
        """
        if self._session is None:
            return None
        res = self._session.res_info
        if res:
            return res.description
        else:
            return None

    def set_stream(self, column_idx, stream):
        if len(self._session.res_info.columns) <= column_idx or column_idx < 0:
            raise ValueError('Invalid value for column_idx')
        self._session.res_info.columns[column_idx].serializer.set_chunk_handler(pytds.tds_types._StreamChunkedHandler(stream))

    @property
    def messages(self):
        """ Messages generated by server, see http://legacy.python.org/dev/peps/pep-0249/#cursor-messages
        """
        if self._session:
            result = []
            for msg in self._session.messages:
                ex = _create_exception_by_message(msg)
                result.append((type(ex), ex))
            return result
        else:
            return None

    @property
    def native_description(self):
        """ todo document
        """
        if self._session is None:
            return None
        res = self._session.res_info
        if res:
            return res.native_descr
        else:
            return None

    def fetchone(self):
        """ Fetches next row, or ``None`` if there are no more rows
        """
        row = self._session.fetchone()
        if row:
            return self._row_factory(row)

    def fetchmany(self, size=None):
        """ Fetches next multiple rows

        :param size: Maximum number of rows to return, default value is cursor.arraysize
        :returns: List of rows
        """
        if size is None:
            size = self.arraysize

        rows = []
        for _ in xrange(size):
            row = self.fetchone()
            if not row:
                break
            rows.append(row)
        return rows

    def fetchall(self):
        """ Fetches all remaining rows
        """
        return list(row for row in self)

    def __next__(self):
        row = self.fetchone()
        if row is None:
            raise StopIteration
        return row

    @staticmethod
    def setinputsizes(sizes=None):
        """
        This method does nothing, as permitted by DB-API specification.
        """
        pass

    @staticmethod
    def setoutputsize(size=None, column=0):
        """
        This method does nothing, as permitted by DB-API specification.
        """
        pass

    def copy_to(self, file=None, table_or_view=None, sep='\t', columns=None,
                check_constraints=False, fire_triggers=False, keep_nulls=False,
                kb_per_batch=None, rows_per_batch=None, order=None, tablock=False,
                schema=None, null_string=None, data=None):
        """ *Experimental*. Efficiently load data to database from file using ``BULK INSERT`` operation

        :param file: Source file-like object, should be in csv format. Specify
          either this or data, not both.
        :param table_or_view: Destination table or view in the database
        :type table_or_view: str

        Optional parameters:

        :keyword sep: Separator used in csv file
        :type sep: str
        :keyword columns: List of Column objects or column names in target
          table to insert to. SQL Server will do some conversions, so these
          may not have to match the actual table definition exactly.
          If not provided will insert into all columns assuming nvarchar(4000)
          NULL for all columns.
          If only the column name is provided, the type is assumed to be
          nvarchar(4000) NULL.
          If rows are given with file, you cannot specify non-string data
          types.
          If rows are given with data, the values must be a type supported by
          the serializer for the column in tds_types.
        :type columns: list
        :keyword check_constraints: Check table constraints for incoming data
        :type check_constraints: bool
        :keyword fire_triggers: Enable or disable triggers for table
        :type fire_triggers: bool
        :keyword keep_nulls: If enabled null values inserted as-is, instead of
          inserting default value for column
        :type keep_nulls: bool
        :keyword kb_per_batch: Kilobytes per batch can be used to optimize performance, see MSSQL
          server documentation for details
        :type kb_per_batch: int
        :keyword rows_per_batch: Rows per batch can be used to optimize performance, see MSSQL
          server documentation for details
        :type rows_per_batch: int
        :keyword order: The ordering of the data in source table. List of columns with ASC or DESC suffix.
          E.g. ``['order_id ASC', 'name DESC']``
          Can be used to optimize performance, see MSSQL server documentation for details
        :type order: list
        :keyword tablock: Enable or disable table lock for the duration of bulk load
        :keyword schema: Name of schema for table or view, if not specified default schema will be used
        :keyword null_string: String that should be interpreted as a NULL when
          reading the CSV file. Has no meaning if using data instead of file.
        :keyword data: The data to insert as an iterable of rows, which are
          iterables of values. Specify either this or file, not both.
        """
        conn = self._conn()
        rows = None
        if data is None:
            import csv
            reader = csv.reader(file, delimiter=sep)

            if null_string is not None:
                def _convert_null_strings(csv_reader):
                    for row in csv_reader:
                        yield [r if r != null_string else None for r in row]

                reader = _convert_null_strings(reader)

            rows = reader
        else:
            rows = data

        obj_name = tds_base.tds_quote_id(table_or_view)
        if schema:
            obj_name = '{0}.{1}'.format(tds_base.tds_quote_id(schema), obj_name)
        if columns:
            metadata = []
            for column in columns:
                if isinstance(column, Column):
                    metadata.append(column)
                else:
                    metadata.append(Column(name=column, type=NVarCharType(size=4000), flags=Column.fNullable))
        else:
            self.execute('select top 1 * from {} where 1<>1'.format(obj_name))
            metadata = [Column(name=col[0], type=NVarCharType(size=4000), flags=Column.fNullable if col[6] else 0)
                        for col in self.description]
        col_defs = ','.join('{0} {1}'.format(tds_base.tds_quote_id(col.column_name), col.type.get_declaration())
                            for col in metadata)
        with_opts = []
        if check_constraints:
            with_opts.append('CHECK_CONSTRAINTS')
        if fire_triggers:
            with_opts.append('FIRE_TRIGGERS')
        if keep_nulls:
            with_opts.append('KEEP_NULLS')
        if kb_per_batch:
            with_opts.append('KILOBYTES_PER_BATCH = {0}'.format(kb_per_batch))
        if rows_per_batch:
            with_opts.append('ROWS_PER_BATCH = {0}'.format(rows_per_batch))
        if order:
            with_opts.append('ORDER({0})'.format(','.join(order)))
        if tablock:
            with_opts.append('TABLOCK')
        with_part = ''
        if with_opts:
            with_part = 'WITH ({0})'.format(','.join(with_opts))
        operation = 'INSERT BULK {0}({1}) {2}'.format(obj_name, col_defs, with_part)
        self.execute(operation)
        self._session.submit_bulk(metadata, rows)
        self._session.process_simple_request()


class _MarsCursor(Cursor):
    def _assert_open(self):
        conn = self._conn
        if conn is not None:
            conn = conn()
        if not conn:
            raise InterfaceError('Cursor is closed')
        conn._assert_open()
        if not self._session.is_connected():
            self._session = conn._conn.create_session(self._tzinfo_factory)
        return conn

    @property
    def spid(self):
        # not thread safe for connection
        conn = self._assert_open()
        dirty = conn._dirty
        spid = self.execute_scalar('select @@SPID')
        conn._dirty = dirty
        return spid

    def cancel(self):
        self._assert_open()
        self._session.cancel_if_pending()

    def close(self):
        """
        Closes the cursor. The cursor is unusable from this point.
        """
        if self._session is not None:
            try:
                self._session.close()
                self._session = None
            except socket.error as e:
                if e.errno != errno.ECONNRESET:
                    raise

    def execute(self, operation, params=()):
        self._assert_open()
        self._execute(operation, params)
        # for compatibility with pyodbc
        return self

    def callproc(self, procname, parameters=()):
        """
        Call a stored procedure with the given name.

        :param procname: The name of the procedure to call
        :type procname: str
        :keyword parameters: The optional parameters for the procedure
        :type parameters: sequence
        """
        self._assert_open()
        return self._callproc(procname, parameters)

    def _begin_tran(self, isolation_level):
        self._assert_open()
        self._session.begin_tran(isolation_level=isolation_level)

    def _commit(self, cont, isolation_level=0):
        conn = self._assert_open()
        self._session.commit(cont=cont, isolation_level=isolation_level)
        conn._dirty = False

    def _rollback(self, cont, isolation_level=0):
        conn = self._assert_open()
        self._session.rollback(cont=cont, isolation_level=isolation_level)
        conn._dirty = False


def _resolve_instance_port(server, port, instance, timeout=5):
    if instance and not port:
        logger.info('querying %s for list of instances', server)
        instances = tds7_get_instances(server, timeout=timeout)
        if instance not in instances:
            raise LoginError("Instance {0} not found on server {1}".format(instance, server))
        instdict = instances[instance]
        if 'tcp' not in instdict:
            raise LoginError("Instance {0} doen't have tcp connections enabled".format(instance))
        port = int(instdict['tcp'])
    return port or 1433


def _parse_server(server):
    instance = ""
    if "\\" in server:
        server, instance = server.split("\\")

    # support MS methods of connecting locally
    if server in (".", "(local)"):
        server = "localhost"

    return server, instance.upper()


# map to servers deques, used to store active/passive servers
# between calls to connect function
# deques are used because they can be rotated
_servers_deques = {}


def _get_servers_deque(servers, database):
    """ Returns deque of servers for given tuple of servers and
    database name.
    This deque have active server at the begining, if first server
    is not accessible at the moment the deque will be rotated,
    second server will be moved to the first position, thirt to the
    second position etc, and previously first server will be moved
    to the last position.
    This allows to remember last successful server between calls
    to connect function.
    """
    key = (servers, database)
    if key not in _servers_deques:
        _servers_deques[key] = deque(servers)
    return _servers_deques[key]


def _parse_connection_string(connstr):
    """
    MSSQL style connection string parser

    Returns normalized dictionary of connection string parameters
    """
    res = {}
    for item in connstr.split(';'):
        item = item.strip()
        if not item:
            continue
        key, value = item.split('=', 1)
        key = key.strip().lower().replace(' ', '_')
        value = value.strip()
        res[key] = value
    return res


def connect(dsn=None, database=None, user=None, password=None, timeout=None,
            login_timeout=15, as_dict=None,
            appname=None, port=None, tds_version=tds_base.TDS74,
            autocommit=False,
            blocksize=4096, use_mars=False, auth=None, readonly=False,
            load_balancer=None, use_tz=None, bytes_to_unicode=True,
            row_strategy=None, failover_partner=None, server=None,
            cafile=None, sock=None, validate_host=True,
            enc_login_only=False, disable_connect_retry=False,
            pooling=False,
            use_sso=False,
            ):
    """
    Opens connection to the database

    :keyword dsn: SQL server host and instance: <host>[\\<instance>]
    :type dsn: string
    :keyword failover_partner: secondary database host, used if primary is not accessible
    :type failover_partner: string
    :keyword database: the database to initially connect to
    :type database: string
    :keyword user: database user to connect as
    :type user: string
    :keyword password: user's password
    :type password: string
    :keyword timeout: query timeout in seconds, default 0 (no timeout)
    :type timeout: int
    :keyword login_timeout: timeout for connection and login in seconds, default 15
    :type login_timeout: int
    :keyword as_dict: whether rows should be returned as dictionaries instead of tuples.
    :type as_dict: boolean
    :keyword appname: Set the application name to use for the connection
    :type appname: string
    :keyword port: the TCP port to use to connect to the server
    :type port: int
    :keyword tds_version: Maximum TDS version to use, should only be used for testing
    :type tds_version: int
    :keyword autocommit: Enable or disable database level autocommit
    :type autocommit: bool
    :keyword blocksize: Size of block for the TDS protocol, usually should not be used
    :type blocksize: int
    :keyword use_mars: Enable or disable MARS
    :type use_mars: bool
    :keyword auth: An instance of authentication method class, e.g. Ntlm or Sspi
    :keyword readonly: Allows to enable read-only mode for connection, only supported by MSSQL 2012,
      earlier versions will ignore this parameter
    :type readonly: bool
    :keyword load_balancer: An instance of load balancer class to use, if not provided will not use load balancer
    :keyword use_tz: Provides timezone for naive database times, if not provided date and time will be returned
      in naive format
    :keyword bytes_to_unicode: If true single byte database strings will be converted to unicode Python strings,
      otherwise will return strings as ``bytes`` without conversion.
    :type bytes_to_unicode: bool
    :keyword row_strategy: strategy used to create rows, determines type of returned rows, can be custom or one of:
      :func:`tuple_row_strategy`, :func:`list_row_strategy`, :func:`dict_row_strategy`,
      :func:`namedtuple_row_strategy`, :func:`recordtype_row_strategy`
    :type row_strategy: function of list of column names returning row factory
    :keyword cafile: Name of the file containing trusted CAs in PEM format, if provided will enable TLS
    :type cafile: str
    :keyword validate_host: Host name validation during TLS connection is enabled by default, if you disable it you
      will be vulnerable to MitM type of attack.
    :type validate_host: bool
    :keyword enc_login_only: Allows you to scope TLS encryption only to an authentication portion.  This means that
      anyone who can observe traffic on your network will be able to see all your SQL requests and potentially modify
      them.
    :type enc_login_only: bool
    :keyword use_sso: Enables SSO login, e.g. Kerberos using SSPI on Windows and kerberos package on other platforms.
             Cannot be used together with auth parameter.
    :returns: An instance of :class:`Connection`
    """
    if use_sso and auth:
        raise ValueError('use_sso cannot be used with auth parameter defined')
    login = _TdsLogin()
    login.client_host_name = socket.gethostname()[:128]
    login.library = "Python TDS Library"
    login.user_name = user or ''
    login.password = password or ''
    login.app_name = appname or 'pytds'
    login.port = port
    login.language = ''  # use database default
    login.attach_db_file = ''
    login.tds_version = tds_version
    if tds_version < tds_base.TDS70:
        raise ValueError('This TDS version is not supported')
    login.database = database or ''
    login.bulk_copy = False
    login.client_lcid = lcid.LANGID_ENGLISH_US
    login.use_mars = use_mars
    login.pid = os.getpid()
    login.change_password = ''
    login.client_id = uuid.getnode()  # client mac address
    login.cafile = cafile
    login.validate_host = validate_host
    login.enc_login_only = enc_login_only
    if cafile:
        if not tls.OPENSSL_AVAILABLE:
            raise ValueError("You are trying to use encryption but pyOpenSSL does not work, you probably "
                             "need to install it first")
        login.tls_ctx = tls.create_context(cafile)
        if login.enc_login_only:
            login.enc_flag = PreLoginEnc.ENCRYPT_OFF
        else:
            login.enc_flag = PreLoginEnc.ENCRYPT_ON
    else:
        login.tls_ctx = None
        login.enc_flag = PreLoginEnc.ENCRYPT_NOT_SUP

    if use_tz:
        login.client_tz = use_tz
    else:
        login.client_tz = pytds.tz.local

    # that will set:
    # ANSI_DEFAULTS to ON,
    # IMPLICIT_TRANSACTIONS to OFF,
    # TEXTSIZE to 0x7FFFFFFF (2GB) (TDS 7.2 and below), TEXTSIZE to infinite (introduced in TDS 7.3),
    # and ROWCOUNT to infinite
    login.option_flag2 = tds_base.TDS_ODBC_ON

    login.connect_timeout = login_timeout
    login.query_timeout = timeout
    login.blocksize = blocksize
    login.readonly = readonly
    login.load_balancer = load_balancer
    login.bytes_to_unicode = bytes_to_unicode

    if server and dsn:
        raise ValueError("Both server and dsn shouldn't be specified")

    if server:
        warnings.warn("server parameter is deprecated, use dsn instead", DeprecationWarning)
        dsn = server

    if load_balancer and failover_partner:
        raise ValueError("Both load_balancer and failover_partner shoudln't be specified")
    if load_balancer:
        servers = [(srv, None) for srv in load_balancer.choose()]
    else:
        servers = [(dsn or 'localhost', port)]
        if failover_partner:
            servers.append((failover_partner, port))

    parsed_servers = []
    for srv, port in servers:
        host, instance = _parse_server(srv)
        if instance and port:
            raise ValueError("Both instance and port shouldn't be specified")
        parsed_servers.append((host, port, instance))

    if use_sso:
        spn = "MSSQLSvc@{}:{}".format(parsed_servers[0][0], parsed_servers[0][1])
        from . import login as pytds_login
        try:
            login.auth = pytds_login.SspiAuth(spn=spn)
        except ImportError:
            login.auth = pytds_login.KerberosAuth(spn)
    else:
        login.auth = auth

    login.servers = _get_servers_deque(tuple(parsed_servers), database)

    # unique connection identifier used to pool connection
    key = (
        dsn,
        login.user_name,
        login.app_name,
        login.tds_version,
        login.database,
        login.client_lcid,
        login.use_mars,
        login.cafile,
        login.blocksize,
        login.readonly,
        login.bytes_to_unicode,
        login.auth,
        login.client_tz,
        autocommit,
    )

    conn = Connection()
    conn._use_tz = use_tz
    conn._autocommit = autocommit
    conn._login = login
    conn._pooling = pooling
    conn._key = key

    assert row_strategy is None or as_dict is None,\
        'Both row_startegy and as_dict were specified, you should use either one or another'
    if as_dict is not None:
        conn.as_dict = as_dict
    elif row_strategy is not None:
        conn._row_strategy = row_strategy
    else:
        conn._row_strategy = tuple_row_strategy # default row strategy

    conn._isolation_level = 0
    conn._dirty = False
    from .tz import FixedOffsetTimezone
    conn._tzinfo_factory = None if use_tz is None else FixedOffsetTimezone
    if disable_connect_retry:
        conn._try_open(timeout=login.connect_timeout, sock=sock)
    else:
        conn._open(sock=sock)
    return conn


def Date(year, month, day):
    return datetime.date(year, month, day)


def DateFromTicks(ticks):
    return datetime.date.fromtimestamp(ticks)


def Time(hour, minute, second, microsecond=0, tzinfo=None):
    return datetime.time(hour, minute, second, microsecond, tzinfo)


def TimeFromTicks(ticks):
    import time
    return Time(*time.localtime(ticks)[3:6])


def Timestamp(year, month, day, hour, minute, second, microseconds=0, tzinfo=None):
    return datetime.datetime(year, month, day, hour, minute, second, microseconds, tzinfo)


def TimestampFromTicks(ticks):
    return datetime.datetime.fromtimestamp(ticks)



import time as _time
from datetime import tzinfo, timedelta

ZERO = timedelta(0)
HOUR = timedelta(hours=1)


# A class building tzinfo objects for fixed-offset time zones.
# Note that FixedOffset(0, "UTC") is a different way to build a
# UTC tzinfo object.


class FixedOffsetTimezone(tzinfo):
    """Fixed offset in minutes east from UTC."""

    def __init__(self, offset, name=None):
        self.__offset = timedelta(minutes=offset)
        self.__name = name

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return self.__name

    def dst(self, dt):
        return ZERO


utc = FixedOffsetTimezone(offset=0, name='UTC')


STDOFFSET = timedelta(seconds=-_time.timezone)
if _time.daylight:
    DSTOFFSET = timedelta(seconds=-_time.altzone)
else:
    DSTOFFSET = STDOFFSET

DSTDIFF = DSTOFFSET - STDOFFSET


class LocalTimezone(tzinfo):

    def utcoffset(self, dt):
        if self._isdst(dt):
            return DSTOFFSET
        else:
            return STDOFFSET

    def dst(self, dt):
        if self._isdst(dt):
            return DSTDIFF
        else:
            return ZERO

    def tzname(self, dt):
        return _time.tzname[self._isdst(dt)]

    def _isdst(self, dt):
        tt = (dt.year, dt.month, dt.day,
              dt.hour, dt.minute, dt.second,
              dt.weekday(), 0, 0)
        stamp = _time.mktime(tt)
        tt = _time.localtime(stamp)
        return tt.tm_isdst > 0

local = LocalTimezone()



# This file implements Session Multiplex Protocol used by MARS connections
# Protocol documentation https://msdn.microsoft.com/en-us/library/cc219643.aspx
import struct
import logging
import threading
import socket
import errno
from six.moves import range
try:
    from bitarray import bitarray
except ImportError:
    class BitArray(list):
        def __init__(self, size):
            super(BitArray, self).__init__()
            self[:] = [False] * size

        def setall(self, val):
            for i in range(len(self)):
                self[i] = val

    bitarray = BitArray
from .tds_base import Error, skipall


logger = logging.getLogger(__name__)


SMP_HEADER = struct.Struct('<BBHLLL')
SMP_ID = 0x53


class _SmpSession(object):
    def __init__(self, mgr, session_id):
        self.session_id = session_id
        self.seq_num_for_send = 0
        self.high_water_for_send = 4
        self._seq_num_for_recv = 0
        self.high_water_for_recv = 4
        self._last_high_water_for_recv = 4
        self._mgr = mgr
        self.recv_queue = []
        self.send_queue = []
        self._state = None
        self._curr_buf_pos = 0
        self._curr_buf = b''

    def __repr__(self):
        fmt = "<_SmpSession sid={} state={} recv_queue={} send_queue={} seq_num_for_send={}>"
        return fmt.format(self.session_id, SessionState.to_str(self._state), self.recv_queue, self.send_queue,
                          self.seq_num_for_send)

    def get_state(self):
        return self._state

    def close(self):
        self._mgr.close_smp_session(self)

    def sendall(self, data):
        self._mgr.send_packet(self, data)

    def _recv_internal(self, size):
        if not self._curr_buf[self._curr_buf_pos:]:
            self._curr_buf = self._mgr.recv_packet(self)
            self._curr_buf_pos = 0
            if not self._curr_buf:
                return 0, 0
        to_read = min(size, len(self._curr_buf) - self._curr_buf_pos)
        offset = self._curr_buf_pos
        self._curr_buf_pos += to_read
        return offset, to_read

    def recv_into(self, buffer, size=0):
        if size == 0:
            size = len(buffer)

        offset, to_read = self._recv_internal(size)
        buffer[:to_read] = self._curr_buf[offset:offset + to_read]
        return to_read

    #def recv(self, size):
    #    offset, to_read = self._recv_internal(size)
    #    return self._curr_buf[offset:offset + to_read]

    def is_connected(self):
        return self._state == SessionState.SESSION_ESTABLISHED


class PacketTypes:
    SYN = 0x1
    ACK = 0x2
    FIN = 0x4
    DATA = 0x8

    #@staticmethod
    #def type_to_str(t):
    #    if t == PacketTypes.SYN:
    #        return 'SYN'
    #    elif t == PacketTypes.ACK:
    #        return 'ACK'
    #    elif t == PacketTypes.DATA:
    #        return 'DATA'
    #    elif t == PacketTypes.FIN:
    #        return 'FIN'


class SessionState:
    SESSION_ESTABLISHED = 1
    CLOSED = 2
    FIN_SENT = 3
    FIN_RECEIVED = 4

    @staticmethod
    def to_str(st):
        if st == SessionState.SESSION_ESTABLISHED:
            return 'SESSION ESTABLISHED'
        elif st == SessionState.CLOSED:
            return 'CLOSED'
        elif st == SessionState.FIN_SENT:
            return 'FIN SENT'
        elif st == SessionState.FIN_RECEIVED:
            return 'FIN RECEIVED'


class SmpManager(object):
    def __init__(self, transport, max_sessions=2 ** 16):
        self._transport = transport
        self._sessions = {}
        self._used_ids_ba = bitarray(max_sessions)
        self._used_ids_ba.setall(False)
        self._lock = threading.RLock()
        self._hdr_buf = memoryview(bytearray(b'\x00' * SMP_HEADER.size))

    def __repr__(self):
        return "<SmpManager sessions={}>".format(self._sessions)

    def create_session(self):
        try:
            session_id = self._used_ids_ba.index(False)
        except ValueError:
            raise Error("Can't create more MARS sessions, close some sessions and try again")
        session = _SmpSession(self, session_id)
        with self._lock:
            self._sessions[session_id] = session
            self._used_ids_ba[session_id] = True
            hdr = SMP_HEADER.pack(
                SMP_ID,
                PacketTypes.SYN,
                session_id,
                SMP_HEADER.size,
                0,
                session.high_water_for_recv,
                )
            self._transport.sendall(hdr)
            session._state = SessionState.SESSION_ESTABLISHED
        return session

    def close_smp_session(self, session):
        if session._state in (SessionState.CLOSED, SessionState.FIN_SENT):
            return
        elif session._state == SessionState.SESSION_ESTABLISHED:
            with self._lock:
                hdr = SMP_HEADER.pack(
                    SMP_ID,
                    PacketTypes.FIN,
                    session.session_id,
                    SMP_HEADER.size,
                    session.seq_num_for_send,
                    session.high_water_for_recv,
                    )
                session._state = SessionState.FIN_SENT
                try:
                    self._transport.sendall(hdr)
                    self.recv_packet(session)
                except (socket.error, OSError) as ex:
                    if ex.errno in (errno.ECONNRESET, errno.EPIPE):
                        session._state = SessionState.CLOSED
                    else:
                        raise ex

    def send_queued_packets(self, session):
        with self._lock:
            while session.send_queue and session.seq_num_for_send < session.high_water_for_send:
                data = session.send_queue.pop(0)
                self.send_packet(session, data)

    @staticmethod
    def _add_one_wrap(val):
        return 0 if val == 2 ** 32 - 1 else val + 1

    def send_packet(self, session, data):
        with self._lock:
            if session.seq_num_for_send < session.high_water_for_send:
                l = SMP_HEADER.size + len(data)
                seq_num = self._add_one_wrap(session.seq_num_for_send)
                hdr = SMP_HEADER.pack(
                    SMP_ID,
                    PacketTypes.DATA,
                    session.session_id,
                    l,
                    seq_num,
                    session.high_water_for_recv,
                    )
                session._last_high_water_for_recv = session.high_water_for_recv
                self._transport.sendall(hdr + data)
                session.seq_num_for_send = self._add_one_wrap(session.seq_num_for_send)
            else:
                session.send_queue.append(data)
                self._read_smp_message()

    def recv_packet(self, session):
        with self._lock:
            if session._state == SessionState.CLOSED:
                return b''
            while not session.recv_queue:
                self._read_smp_message()
                if session._state in (SessionState.CLOSED, SessionState.FIN_RECEIVED):
                    return b''
            session.high_water_for_recv = self._add_one_wrap(session.high_water_for_recv)
            if session.high_water_for_recv - session._last_high_water_for_recv >= 2:
                hdr = SMP_HEADER.pack(
                    SMP_ID,
                    PacketTypes.ACK,
                    session.session_id,
                    SMP_HEADER.size,
                    session.seq_num_for_send,
                    session.high_water_for_recv,
                    )
                self._transport.sendall(hdr)
                session._last_high_water_for_recv = session.high_water_for_recv
            return session.recv_queue.pop(0)

    def _bad_stm(self, message):
        self.close()
        raise Error(message)

    def _read_smp_message(self):
        # caller should acquire lock before calling this function
        buf_pos = 0
        while buf_pos < SMP_HEADER.size:
            read = self._transport.recv_into(self._hdr_buf[buf_pos:])
            buf_pos += read
            if read == 0:
                self._bad_stm('Unexpected EOF while reading SMP header')
        smid, flags, sid, l, seq_num, wnd = SMP_HEADER.unpack(self._hdr_buf)
        if smid != SMP_ID:
            self._bad_stm('Invalid SMP packet signature')
        try:
            session = self._sessions[sid]
        except KeyError:
            self._bad_stm('Invalid SMP packet session id')
        if wnd < session.high_water_for_send:
            self._bad_stm('Invalid WNDW in packet from server')
        if seq_num > session.high_water_for_recv:
            self._bad_stm('Invalid SEQNUM in packet from server')
        if l < SMP_HEADER.size:
            self._bad_stm('Invalid LENGTH in packet from server')
        session._last_recv_seq_num = seq_num
        if flags == PacketTypes.DATA:
            if session._state == SessionState.SESSION_ESTABLISHED:
                if seq_num != self._add_one_wrap(session._seq_num_for_recv):
                    self._bad_stm('Invalid SEQNUM in DATA packet from server')
                session._seq_num_for_recv = seq_num
                remains = l - SMP_HEADER.size
                while remains:
                    data = self._transport.recv(remains)
                    session.recv_queue.append(data)
                    remains -= len(data)
                if wnd > session.high_water_for_send:
                    session.high_water_for_send = wnd
                    self.send_queued_packets(session)

            elif session._state == SessionState.FIN_SENT:
                skipall(self._transport, l - SMP_HEADER.size)
            else:
                self._bad_stm('Unexpected DATA packet from server')
        elif flags == PacketTypes.ACK:
            if session._state in (SessionState.FIN_RECEIVED, SessionState.CLOSED):
                self._bad_stm('Unexpected ACK packet from server')
            if seq_num != session._seq_num_for_recv:
                self._bad_stm('Invalid SEQNUM in ACK packet from server')
            session.high_water_for_send = wnd
            self.send_queued_packets(session)
        elif flags == PacketTypes.FIN:
            assert session._state in (SessionState.SESSION_ESTABLISHED, SessionState.FIN_SENT, SessionState.FIN_RECEIVED)
            if session._state == SessionState.SESSION_ESTABLISHED:
                session._state = SessionState.FIN_RECEIVED
            elif session._state == SessionState.FIN_SENT:
                session._state = SessionState.CLOSED
                del self._sessions[session.session_id]
                self._used_ids_ba[session.session_id] = False
            elif session._state == SessionState.FIN_RECEIVED:
                self._bad_stm('Unexpected FIN packet from server')
        elif flags == PacketTypes.SYN:
            self._bad_stm('Unexpected SYN packet from server')
        else:
            self._bad_stm('Unexpected FLAGS in packet from server')

    def close(self):
        self._transport.close()

    def transport_closed(self):
        for session in self._sessions.values():
            session._state = SessionState.CLOSED



import struct
import cProfile
import pstats
import pytds.tds


BUFSIZE = 4096
HEADER = struct.Struct('>BBHHBx')


class Sock:
    def __init__(self):
        self._read_pos = 0
        self._buf = bytearray(b'\x00' * BUFSIZE)
        HEADER.pack_into(self._buf, 0, 0, 0, BUFSIZE, 0, 0)

    def sendall(self, data, flags=0):
        pass

    def recv_into(self, buffer, size=0):
        if size == 0:
            size = len(buffer)
        if self._read_pos >= BUFSIZE:
            HEADER.pack_into(self._buf, 0, 0, 0, BUFSIZE, 0, 0)
            self._read_pos = 0
        to_read = min(size, BUFSIZE - self._read_pos)
        buffer[:to_read] = self._buf[self._read_pos:self._read_pos+to_read]
        return to_read

    def recv(self, size):
        if self._read_pos >= len(self._buf):
            HEADER.pack_into(self._buf, 0, 0, 0, BUFSIZE, 0, 0)
            self._read_pos = 0
        res = self._buf[self._read_pos:self._read_pos + size]
        self._read_pos += len(res)
        return res

    def close(self):
        pass


class Session:
    def __init__(self):
        self._transport = Sock()


sess = Session()

rdr = pytds.tds._TdsReader(sess)
pr = cProfile.Profile()
pr.enable()
for _ in range(50000):
    rdr.recv(BUFSIZE)
pr.disable()
sortby = 'tottime'
ps = pstats.Stats(pr).sort_stats(sortby)
ps.print_stats()



try:
    import unittest2 as unittest
except:
    import unittest
import ctypes
from ctypes import create_string_buffer
import settings
import socket
import sys

@unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
class SspiTest(unittest.TestCase):
    def test_enum_security_packages(self):
        import pytds.sspi
        pytds.sspi.enum_security_packages()

    def test_credentials(self):
        import pytds.sspi
        cred = pytds.sspi.SspiCredentials('Negotiate', pytds.sspi.SECPKG_CRED_OUTBOUND)
        cred.query_user_name()
        cred.close()

    def test_make_buffers(self):
        import pytds.sspi
        buf = create_string_buffer(1000)
        bufs = [(pytds.sspi.SECBUFFER_TOKEN, buf)]
        desc = pytds.sspi._make_buffers_desc(bufs)
        self.assertEqual(desc.ulVersion, pytds.sspi.SECBUFFER_VERSION)
        self.assertEqual(desc.cBuffers, len(bufs))
        self.assertEqual(desc.pBuffers[0].cbBuffer, len(bufs[0][1]))
        self.assertEqual(desc.pBuffers[0].BufferType, bufs[0][0])
        self.assertEqual(desc.pBuffers[0].pvBuffer, ctypes.cast(bufs[0][1], pytds.sspi.PVOID).value)

    def test_sec_context(self):
        import pytds.sspi
        cred = pytds.sspi.SspiCredentials(
            'Negotiate',
            pytds.sspi.SECPKG_CRED_OUTBOUND)

        token_buf = create_string_buffer(10000)
        bufs = [(pytds.sspi.SECBUFFER_TOKEN, token_buf)]
        server = settings.HOST
        if '\\' in server:
            server, _ = server.split('\\')
        host, _, _ = socket.gethostbyname_ex(server)
        target_name = 'MSSQLSvc/{0}:1433'.format(host)
        ctx, status, bufs = cred.create_context(
            flags=pytds.sspi.ISC_REQ_CONFIDENTIALITY|pytds.sspi.ISC_REQ_REPLAY_DETECT|pytds.sspi.ISC_REQ_CONNECTION,
            byte_ordering='network',
            target_name=target_name,
            output_buffers=bufs)
        if status == pytds.sspi.Status.SEC_I_COMPLETE_AND_CONTINUE or status == pytds.sspi.Status.SEC_I_CONTINUE_NEEDED:
            ctx.complete_auth_token(bufs)

        #realbuf = create_string_buffer(10000)
        #buf = SecBuffer()
        #buf.cbBuffer = len(realbuf)
        #buf.BufferType = SECBUFFER_TOKEN
        #buf.pvBuffer = cast(realbuf, PVOID)
        #bufs = SecBufferDesc()
        #bufs.ulVersion = SECBUFFER_VERSION
        #bufs.cBuffers = 1
        #bufs.pBuffers = pointer(buf)
        #byte_ordering = 'network'
        #output_buffers = bufs
        #from pytds.sspi import _SecContext
        #ctx = _SecContext()
        #ctx._handle = SecHandle()
        #ctx._ts = TimeStamp()
        #ctx._attrs = ULONG()
        #status = sec_fn.InitializeSecurityContext(
        #        ctypes.byref(cred._handle),
        #        None,
        #        'MSSQLSvc/misha-pc:1433',
        #        ISC_REQ_CONNECTION,
        #        0,
        #        SECURITY_NETWORK_DREP if byte_ordering == 'network' else SECURITY_NATIVE_DREP,
        #        None,
        #        0,
        #        byref(ctx._handle),
        #        byref(bufs),
        #        byref(ctx._attrs),
        #        byref(ctx._ts));
        #pass



import os
import setuptools
from setuptools import setup
import version

requirements = list(open(os.path.join(os.path.dirname(__file__), 'requirements.txt'), 'r').readlines())

print(setuptools.find_packages('src'))

setup(name='python-tds',
      version=version.get_git_version(),
      description='Python DBAPI driver for MSSQL using pure Python TDS (Tabular Data Stream) protocol implementation',
      author='Mikhail Denisenko',
      author_email='denisenkom@gmail.com',
      url='https://github.com/denisenkom/pytds',
      license="MIT",
      packages=['pytds'],
      package_dir={'': 'src'},
      classifiers=[
          'Development Status :: 4 - Beta',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
      ],
      zip_safe=True,
      install_requires=requirements,
      )



# Copyright 2010 Michael Murr
#
# This file is part of LibForensics.
#
# LibForensics is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# LibForensics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with LibForensics.  If not, see <http://www.gnu.org/licenses/>.

"""Constants for Locale IDs. (LCIDs)"""

__docformat__ = "restructuredtext en"
__all__ = [
    "LANGID_AFRIKAANS", "LANGID_ALBANIAN", "LANGID_AMHARIC", "LANGID_ARABIC",
    "LANGID_ARABIC_ALGERIA", "LANGID_ARABIC_BAHRAIN", "LANGID_ARABIC_EGYPT",
    "LANGID_ARABIC_IRAQ", "LANGID_ARABIC_JORDAN", "LANGID_ARABIC_KUWAIT",
    "LANGID_ARABIC_LEBANON", "LANGID_ARABIC_LIBYA", "LANGID_ARABIC_MOROCCO",
    "LANGID_ARABIC_OMAN", "LANGID_ARABIC_QATAR", "LANGID_ARABIC_SYRIA",
    "LANGID_ARABIC_TUNISIA", "LANGID_ARABIC_UAE", "LANGID_ARABIC_YEMEN",
    "LANGID_ARMENIAN", "LANGID_ASSAMESE", "LANGID_AZERI_CYRILLIC",
    "LANGID_AZERI_LATIN", "LANGID_BASQUE", "LANGID_BELGIAN_DUTCH",
    "LANGID_BELGIAN_FRENCH", "LANGID_BENGALI", "LANGID_BULGARIAN",
    "LANGID_BURMESE", "LANGID_BYELORUSSIAN", "LANGID_CATALAN",
    "LANGID_CHEROKEE", "LANGID_CHINESE_HONG_KONG_SAR",
    "LANGID_CHINESE_MACAO_SAR", "LANGID_CHINESE_SINGAPORE", "LANGID_CROATIAN",
    "LANGID_CZECH", "LANGID_DANISH", "LANGID_DIVEHI", "LANGID_DUTCH",
    "LANGID_EDO", "LANGID_ENGLISH_AUS", "LANGID_ENGLISH_BELIZE",
    "LANGID_ENGLISH_CANADIAN", "LANGID_ENGLISH_CARIBBEAN",
    "LANGID_ENGLISH_INDONESIA", "LANGID_ENGLISH_IRELAND",
    "LANGID_ENGLISH_JAMAICA", "LANGID_ENGLISH_NEW_ZEALAND",
    "LANGID_ENGLISH_PHILIPPINES", "LANGID_ENGLISH_SOUTH_AFRICA",
    "LANGID_ENGLISH_TRINIDAD_TOBAGO", "LANGID_ENGLISH_UK", "LANGID_ENGLISH_US",
    "LANGID_ENGLISH_ZIMBABWE", "LANGID_ESTONIAN", "LANGID_FAEROESE",
    "LANGID_FILIPINO", "LANGID_FINNISH", "LANGID_FRENCH",
    "LANGID_FRENCH_CAMEROON", "LANGID_FRENCH_CANADIAN",
    "LANGID_FRENCH_CONGO_D_R_C", "LANGID_FRENCH_COTED_IVOIRE",
    "LANGID_FRENCH_HAITI", "LANGID_FRENCH_LUXEMBOURG", "LANGID_FRENCH_MALI",
    "LANGID_FRENCH_MONACO", "LANGID_FRENCH_MOROCCO", "LANGID_FRENCH_REUNION",
    "LANGID_FRENCH_SENEGAL", "LANGID_FRENCH_WEST_INDIES",
    "LANGID_FRISIAN_NETHERLANDS", "LANGID_FULFULDE", "LANGID_GAELIC_IRELAND",
    "LANGID_GAELIC_SCOTLAND", "LANGID_GALICIAN", "LANGID_GEORGIAN",
    "LANGID_GERMAN", "LANGID_GERMAN_AUSTRIA", "LANGID_GERMAN_LIECHTENSTEIN",
    "LANGID_GERMAN_LUXEMBOURG", "LANGID_GREEK", "LANGID_GUARANI",
    "LANGID_GUJARATI", "LANGID_HAUSA", "LANGID_HAWAIIAN", "LANGID_HEBREW",
    "LANGID_HINDI", "LANGID_HUNGARIAN", "LANGID_IBIBIO", "LANGID_ICELANDIC",
    "LANGID_IGBO", "LANGID_INDONESIAN", "LANGID_INUKTITUT", "LANGID_ITALIAN",
    "LANGID_JAPANESE", "LANGID_KANNADA", "LANGID_KANURI", "LANGID_KASHMIRI",
    "LANGID_KAZAKH", "LANGID_KHMER", "LANGID_KIRGHIZ", "LANGID_KONKANI",
    "LANGID_KOREAN", "LANGID_KYRGYZ", "LANGID_LANGUAGE_NONE", "LANGID_LAO",
    "LANGID_LATIN", "LANGID_LATVIAN", "LANGID_LITHUANIAN",
    "LANGID_MACEDONIAN_FYROM", "LANGID_MALAYALAM", "LANGID_MALAYSIAN",
    "LANGID_MALAY_BRUNEI_DARUSSALAM", "LANGID_MALTESE", "LANGID_MANIPURI",
    "LANGID_MARATHI", "LANGID_MEXICAN_SPANISH", "LANGID_MONGOLIAN",
    "LANGID_NEPALI", "LANGID_NORWEGIAN_BOKMOL", "LANGID_NORWEGIAN_NYNORSK",
    "LANGID_NO_PROOFING", "LANGID_ORIYA", "LANGID_OROMO", "LANGID_PASHTO",
    "LANGID_PERSIAN", "LANGID_POLISH", "LANGID_PORTUGUESE",
    "LANGID_PORTUGUESE_BRAZIL", "LANGID_PUNJABI", "LANGID_RHAETO_ROMANIC",
    "LANGID_ROMANIAN", "LANGID_ROMANIAN_MOLDOVA", "LANGID_RUSSIAN",
    "LANGID_RUSSIAN_MOLDOVA", "LANGID_SAMI_LAPPISH", "LANGID_SANSKRIT",
    "LANGID_SERBIAN_CYRILLIC", "LANGID_SERBIAN_LATIN", "LANGID_SESOTHO",
    "LANGID_SIMPLIFIED_CHINESE", "LANGID_SINDHI", "LANGID_SINDHI_PAKISTAN",
    "LANGID_SINHALESE", "LANGID_SLOVAK", "LANGID_SLOVENIAN", "LANGID_SOMALI",
    "LANGID_SORBIAN", "LANGID_SPANISH", "LANGID_SPANISH_ARGENTINA",
    "LANGID_SPANISH_BOLIVIA", "LANGID_SPANISH_CHILE",
    "LANGID_SPANISH_COLOMBIA", "LANGID_SPANISH_COSTA_RICA",
    "LANGID_SPANISH_DOMINICAN_REPUBLIC", "LANGID_SPANISH_ECUADOR",
    "LANGID_SPANISH_EL_SALVADOR", "LANGID_SPANISH_GUATEMALA",
    "LANGID_SPANISH_HONDURAS", "LANGID_SPANISH_MODERN_SORT",
    "LANGID_SPANISH_NICARAGUA", "LANGID_SPANISH_PANAMA",
    "LANGID_SPANISH_PARAGUAY", "LANGID_SPANISH_PERU",
    "LANGID_SPANISH_PUERTO_RICO", "LANGID_SPANISH_URUGUAY",
    "LANGID_SPANISH_VENEZUELA", "LANGID_SUTU", "LANGID_SWAHILI",
    "LANGID_SWEDISH", "LANGID_SWEDISH_FINLAND", "LANGID_SWISS_FRENCH",
    "LANGID_SWISS_GERMAN", "LANGID_SWISS_ITALIAN", "LANGID_SYRIAC",
    "LANGID_TAJIK", "LANGID_TAMAZIGHT", "LANGID_TAMAZIGHT_LATIN",
    "LANGID_TAMIL", "LANGID_TATAR", "LANGID_TELUGU", "LANGID_THAI",
    "LANGID_TIBETAN", "LANGID_TIGRIGNA_ERITREA", "LANGID_TIGRIGNA_ETHIOPIC",
    "LANGID_TRADITIONAL_CHINESE", "LANGID_TSONGA", "LANGID_TSWANA",
    "LANGID_TURKISH", "LANGID_TURKMEN", "LANGID_UKRAINIAN", "LANGID_URDU",
    "LANGID_UZBEK_CYRILLIC", "LANGID_UZBEK_LATIN", "LANGID_VENDA",
    "LANGID_VIETNAMESE", "LANGID_WELSH", "LANGID_XHOSA", "LANGID_YI",
    "LANGID_YIDDISH", "LANGID_YORUBA", "LANGID_ZULU",

    "lang_id_names"
]

LANGID_AFRIKAANS = 1078
LANGID_ALBANIAN = 1052
LANGID_AMHARIC = 1118
LANGID_ARABIC = 1025
LANGID_ARABIC_ALGERIA = 5121
LANGID_ARABIC_BAHRAIN = 15361
LANGID_ARABIC_EGYPT = 3073
LANGID_ARABIC_IRAQ = 2049
LANGID_ARABIC_JORDAN = 11265
LANGID_ARABIC_KUWAIT = 13313
LANGID_ARABIC_LEBANON = 12289
LANGID_ARABIC_LIBYA = 4097
LANGID_ARABIC_MOROCCO = 6145
LANGID_ARABIC_OMAN = 8193
LANGID_ARABIC_QATAR = 16385
LANGID_ARABIC_SYRIA = 10241
LANGID_ARABIC_TUNISIA = 7169
LANGID_ARABIC_UAE = 14337
LANGID_ARABIC_YEMEN = 9217
LANGID_ARMENIAN = 1067
LANGID_ASSAMESE = 1101
LANGID_AZERI_CYRILLIC = 2092
LANGID_AZERI_LATIN = 1068
LANGID_BASQUE = 1069
LANGID_BELGIAN_DUTCH = 2067
LANGID_BELGIAN_FRENCH = 2060
LANGID_BENGALI = 1093
LANGID_BULGARIAN = 1026
LANGID_BURMESE = 1109
LANGID_BYELORUSSIAN = 1059
LANGID_CATALAN = 1027
LANGID_CHEROKEE = 1116
LANGID_CHINESE_HONG_KONG_SAR = 3076
LANGID_CHINESE_MACAO_SAR = 5124
LANGID_CHINESE_SINGAPORE = 4100
LANGID_CROATIAN = 1050
LANGID_CZECH = 1029
LANGID_DANISH = 1030
LANGID_DIVEHI = 1125
LANGID_DUTCH = 1043
LANGID_EDO = 1126
LANGID_ENGLISH_AUS = 3081
LANGID_ENGLISH_BELIZE = 10249
LANGID_ENGLISH_CANADIAN = 4105
LANGID_ENGLISH_CARIBBEAN = 9225
LANGID_ENGLISH_INDONESIA = 14345
LANGID_ENGLISH_IRELAND = 6153
LANGID_ENGLISH_JAMAICA = 8201
LANGID_ENGLISH_NEW_ZEALAND = 5129
LANGID_ENGLISH_PHILIPPINES = 13321
LANGID_ENGLISH_SOUTH_AFRICA = 7177
LANGID_ENGLISH_TRINIDAD_TOBAGO = 11273
LANGID_ENGLISH_UK = 2057
LANGID_ENGLISH_US = 1033
LANGID_ENGLISH_ZIMBABWE = 12297
LANGID_ESTONIAN = 1061
LANGID_FAEROESE = 1080
LANGID_FILIPINO = 1124
LANGID_FINNISH = 1035
LANGID_FRENCH = 1036
LANGID_FRENCH_CAMEROON = 11276
LANGID_FRENCH_CANADIAN = 3084
LANGID_FRENCH_CONGO_D_R_C = 9228
LANGID_FRENCH_COTED_IVOIRE = 12300
LANGID_FRENCH_HAITI = 15372
LANGID_FRENCH_LUXEMBOURG = 5132
LANGID_FRENCH_MALI = 13324
LANGID_FRENCH_MONACO = 6156
LANGID_FRENCH_MOROCCO = 14348
LANGID_FRENCH_REUNION = 8204
LANGID_FRENCH_SENEGAL = 10252
LANGID_FRENCH_WEST_INDIES = 7180
LANGID_FRISIAN_NETHERLANDS = 1122
LANGID_FULFULDE = 1127
LANGID_GAELIC_IRELAND = 2108
LANGID_GAELIC_SCOTLAND = 1084
LANGID_GALICIAN = 1110
LANGID_GEORGIAN = 1079
LANGID_GERMAN = 1031
LANGID_GERMAN_AUSTRIA = 3079
LANGID_GERMAN_LIECHTENSTEIN = 5127
LANGID_GERMAN_LUXEMBOURG = 4103
LANGID_GREEK = 1032
LANGID_GUARANI = 1140
LANGID_GUJARATI = 1095
LANGID_HAUSA = 1128
LANGID_HAWAIIAN = 1141
LANGID_HEBREW = 1037
LANGID_HINDI = 1081
LANGID_HUNGARIAN = 1038
LANGID_IBIBIO = 1129
LANGID_ICELANDIC = 1039
LANGID_IGBO = 1136
LANGID_INDONESIAN = 1057
LANGID_INUKTITUT = 1117
LANGID_ITALIAN = 1040
LANGID_JAPANESE = 1041
LANGID_KANNADA = 1099
LANGID_KANURI = 1137
LANGID_KASHMIRI = 1120
LANGID_KAZAKH = 1087
LANGID_KHMER = 1107
LANGID_KIRGHIZ = 1088
LANGID_KONKANI = 1111
LANGID_KOREAN = 1042
LANGID_KYRGYZ = 1088
LANGID_LANGUAGE_NONE = 0
LANGID_LAO = 1108
LANGID_LATIN = 1142
LANGID_LATVIAN = 1062
LANGID_LITHUANIAN = 1063
LANGID_MACEDONIAN_FYROM = 1071
LANGID_MALAYALAM = 1100
LANGID_MALAY_BRUNEI_DARUSSALAM = 2110
LANGID_MALAYSIAN = 1086
LANGID_MALTESE = 1082
LANGID_MANIPURI = 1112
LANGID_MARATHI = 1102
LANGID_MEXICAN_SPANISH = 2058
LANGID_MONGOLIAN = 1104
LANGID_NEPALI = 1121
LANGID_NO_PROOFING = 1024
LANGID_NORWEGIAN_BOKMOL = 1044
LANGID_NORWEGIAN_NYNORSK = 2068
LANGID_ORIYA = 1096
LANGID_OROMO = 1138
LANGID_PASHTO = 1123
LANGID_PERSIAN = 1065
LANGID_POLISH = 1045
LANGID_PORTUGUESE = 2070
LANGID_PORTUGUESE_BRAZIL = 1046
LANGID_PUNJABI = 1094
LANGID_RHAETO_ROMANIC = 1047
LANGID_ROMANIAN = 1048
LANGID_ROMANIAN_MOLDOVA = 2072
LANGID_RUSSIAN = 1049
LANGID_RUSSIAN_MOLDOVA = 2073
LANGID_SAMI_LAPPISH = 1083
LANGID_SANSKRIT = 1103
LANGID_SERBIAN_CYRILLIC = 3098
LANGID_SERBIAN_LATIN = 2074
LANGID_SESOTHO = 1072
LANGID_SIMPLIFIED_CHINESE = 2052
LANGID_SINDHI = 1113
LANGID_SINDHI_PAKISTAN = 2137
LANGID_SINHALESE = 1115
LANGID_SLOVAK = 1051
LANGID_SLOVENIAN = 1060
LANGID_SOMALI = 1143
LANGID_SORBIAN = 1070
LANGID_SPANISH = 1034
LANGID_SPANISH_ARGENTINA = 11274
LANGID_SPANISH_BOLIVIA = 16394
LANGID_SPANISH_CHILE = 13322
LANGID_SPANISH_COLOMBIA = 9226
LANGID_SPANISH_COSTA_RICA = 5130
LANGID_SPANISH_DOMINICAN_REPUBLIC = 7178
LANGID_SPANISH_ECUADOR = 12298
LANGID_SPANISH_EL_SALVADOR = 17418
LANGID_SPANISH_GUATEMALA = 4106
LANGID_SPANISH_HONDURAS = 18442
LANGID_SPANISH_MODERN_SORT = 3082
LANGID_SPANISH_NICARAGUA = 19466
LANGID_SPANISH_PANAMA = 6154
LANGID_SPANISH_PARAGUAY = 15370
LANGID_SPANISH_PERU = 10250
LANGID_SPANISH_PUERTO_RICO = 20490
LANGID_SPANISH_URUGUAY = 14346
LANGID_SPANISH_VENEZUELA = 8202
LANGID_SUTU = 1072
LANGID_SWAHILI = 1089
LANGID_SWEDISH = 1053
LANGID_SWEDISH_FINLAND = 2077
LANGID_SWISS_FRENCH = 4108
LANGID_SWISS_GERMAN = 2055
LANGID_SWISS_ITALIAN = 2064
LANGID_SYRIAC = 1114
LANGID_TAJIK = 1064
LANGID_TAMAZIGHT = 1119
LANGID_TAMAZIGHT_LATIN = 2143
LANGID_TAMIL = 1097
LANGID_TATAR = 1092
LANGID_TELUGU = 1098
LANGID_THAI = 1054
LANGID_TIBETAN = 1105
LANGID_TIGRIGNA_ERITREA = 2163
LANGID_TIGRIGNA_ETHIOPIC = 1139
LANGID_TRADITIONAL_CHINESE = 1028
LANGID_TSONGA = 1073
LANGID_TSWANA = 1074
LANGID_TURKISH = 1055
LANGID_TURKMEN = 1090
LANGID_UKRAINIAN = 1058
LANGID_URDU = 1056
LANGID_UZBEK_CYRILLIC = 2115
LANGID_UZBEK_LATIN = 1091
LANGID_VENDA = 1075
LANGID_VIETNAMESE = 1066
LANGID_WELSH = 1106
LANGID_XHOSA = 1076
LANGID_YI = 1144
LANGID_YIDDISH = 1085
LANGID_YORUBA = 1130
LANGID_ZULU = 1077

lang_id_names = {
    LANGID_AFRIKAANS: "African",
    LANGID_ALBANIAN: "Albanian",
    LANGID_AMHARIC: "Amharic",
    LANGID_ARABIC: "Arabic",
    LANGID_ARABIC_ALGERIA: "Arabic Algerian",
    LANGID_ARABIC_BAHRAIN: "Arabic Bahraini",
    LANGID_ARABIC_EGYPT: "Arabic Egyptian",
    LANGID_ARABIC_IRAQ: "Arabic Iraqi",
    LANGID_ARABIC_JORDAN: "Arabic Jordanian",
    LANGID_ARABIC_KUWAIT: "Arabic Kuwaiti",
    LANGID_ARABIC_LEBANON: "Arabic Lebanese",
    LANGID_ARABIC_LIBYA: "Arabic Libyan",
    LANGID_ARABIC_MOROCCO: "Arabic Moroccan",
    LANGID_ARABIC_OMAN: "Arabic Omani",
    LANGID_ARABIC_QATAR: "Arabic Qatari",
    LANGID_ARABIC_SYRIA: "Arabic Syrian",
    LANGID_ARABIC_TUNISIA: "Arabic Tunisian",
    LANGID_ARABIC_UAE: "Arabic United Arab Emirates",
    LANGID_ARABIC_YEMEN: "Arabic Yemeni",
    LANGID_ARMENIAN: "Armenian",
    LANGID_ASSAMESE: "Assamese",
    LANGID_AZERI_CYRILLIC: "Azeri Cyrillic",
    LANGID_AZERI_LATIN: "Azeri Latin",
    LANGID_BASQUE: "Basque",
    LANGID_BELGIAN_DUTCH: "Belgian Dutch",
    LANGID_BELGIAN_FRENCH: "Belgian French",
    LANGID_BENGALI: "Bengali",
    LANGID_BULGARIAN: "Bulgarian",
    LANGID_BURMESE: "Burmese",
    LANGID_BYELORUSSIAN: "Byelorussian",
    LANGID_CATALAN: "Catalan",
    LANGID_CHEROKEE: "Cherokee",
    LANGID_CHINESE_HONG_KONG_SAR: "Chinese Hong Kong SAR",
    LANGID_CHINESE_MACAO_SAR: "Chinese Macao SAR",
    LANGID_CHINESE_SINGAPORE: "Chinese Singapore",
    LANGID_CROATIAN: "Croatian",
    LANGID_CZECH: "Czech",
    LANGID_DANISH: "Danish",
    LANGID_DIVEHI: "Divehi",
    LANGID_DUTCH: "Dutch",
    LANGID_EDO: "Edo",
    LANGID_ENGLISH_AUS: "Australian English",
    LANGID_ENGLISH_BELIZE: "Belize English",
    LANGID_ENGLISH_CANADIAN: "Canadian English",
    LANGID_ENGLISH_CARIBBEAN: "Caribbean English",
    LANGID_ENGLISH_INDONESIA: "Indonesian English",
    LANGID_ENGLISH_IRELAND: "Irish English",
    LANGID_ENGLISH_JAMAICA: "Jamaican English",
    LANGID_ENGLISH_NEW_ZEALAND: "New Zealand English",
    LANGID_ENGLISH_PHILIPPINES: "Filipino English",
    LANGID_ENGLISH_SOUTH_AFRICA: "South African English",
    LANGID_ENGLISH_TRINIDAD_TOBAGO: "Tobago Trinidad English",
    LANGID_ENGLISH_UK: "United Kingdom English",
    LANGID_ENGLISH_US: "United States English",
    LANGID_ENGLISH_ZIMBABWE: "Zimbabwe English",
    LANGID_ESTONIAN: "Estonian",
    LANGID_FAEROESE: "Faeroese",
    LANGID_FILIPINO: "Filipino",
    LANGID_FINNISH: "Finnish",
    LANGID_FRENCH: "French",
    LANGID_FRENCH_CAMEROON: "French Cameroon",
    LANGID_FRENCH_CANADIAN: "French Canadian",
    LANGID_FRENCH_CONGO_D_R_C: "French (Congo (DRC))",
    LANGID_FRENCH_COTED_IVOIRE: "French Cote d'Ivoire",
    LANGID_FRENCH_HAITI: "French Haiti",
    LANGID_FRENCH_LUXEMBOURG: "French Luxembourg",
    LANGID_FRENCH_MALI: "French Mali",
    LANGID_FRENCH_MONACO: "French Monaco",
    LANGID_FRENCH_MOROCCO: "French Morocco",
    LANGID_FRENCH_REUNION: "French Reunion",
    LANGID_FRENCH_SENEGAL: "French Senegal",
    LANGID_FRENCH_WEST_INDIES: "French West Indies",
    LANGID_FRISIAN_NETHERLANDS: "Frisian Netherlands",
    LANGID_FULFULDE: "Fulfulde",
    LANGID_GAELIC_IRELAND: "Gaelic Irish",
    LANGID_GAELIC_SCOTLAND: "Gaelic Scottish",
    LANGID_GALICIAN: "Galician",
    LANGID_GEORGIAN: "Georgian",
    LANGID_GERMAN: "German",
    LANGID_GERMAN_AUSTRIA: "German Austrian",
    LANGID_GERMAN_LIECHTENSTEIN: "German Liechtenstein",
    LANGID_GERMAN_LUXEMBOURG: "German Luxembourg",
    LANGID_GREEK: "Greek",
    LANGID_GUARANI: "Guarani",
    LANGID_GUJARATI: "Gujarati",
    LANGID_HAUSA: "Hausa",
    LANGID_HAWAIIAN: "Hawaiian",
    LANGID_HEBREW: "Hebrew",
    LANGID_HINDI: "Hindi",
    LANGID_HUNGARIAN: "Hungarian",
    LANGID_IBIBIO: "Ibibio",
    LANGID_ICELANDIC: "Icelandic",
    LANGID_IGBO: "Igbo",
    LANGID_INDONESIAN: "Indonesian",
    LANGID_INUKTITUT: "Inuktitut",
    LANGID_ITALIAN: "Italian",
    LANGID_JAPANESE: "Japanese",
    LANGID_KANNADA: "Kannada",
    LANGID_KANURI: "Kanuri",
    LANGID_KASHMIRI: "Kashmiri",
    LANGID_KAZAKH: "Kazakh",
    LANGID_KHMER: "Khmer",
    LANGID_KIRGHIZ: "Kirghiz",
    LANGID_KONKANI: "Konkani",
    LANGID_KOREAN: "Korean",
    LANGID_KYRGYZ: "Kyrgyz",
    LANGID_LANGUAGE_NONE: "No specified",
    LANGID_LAO: "Lao",
    LANGID_LATIN: "Latin",
    LANGID_LATVIAN: "Latvian",
    LANGID_LITHUANIAN: "Lithuanian",
    LANGID_MACEDONIAN_FYROM: "Macedonian (FYROM)",
    LANGID_MALAYALAM: "Malayalam",
    LANGID_MALAY_BRUNEI_DARUSSALAM: "Malay Brunei Darussalam",
    LANGID_MALAYSIAN: "Malaysian",
    LANGID_MALTESE: "Maltese",
    LANGID_MANIPURI: "Manipuri",
    LANGID_MARATHI: "Marathi",
    LANGID_MEXICAN_SPANISH: "Mexican Spanish",
    LANGID_MONGOLIAN: "Mongolian",
    LANGID_NEPALI: "Nepali",
    LANGID_NO_PROOFING: "Disables proofing",
    LANGID_NORWEGIAN_BOKMOL: "Norwegian Bokmol",
    LANGID_NORWEGIAN_NYNORSK: "Norwegian Nynorsk",
    LANGID_ORIYA: "Oriya",
    LANGID_OROMO: "Oromo",
    LANGID_PASHTO: "Pashto",
    LANGID_PERSIAN: "Persian",
    LANGID_POLISH: "Polish",
    LANGID_PORTUGUESE: "Portuguese",
    LANGID_PORTUGUESE_BRAZIL: "Portuguese (Brazil)",
    LANGID_PUNJABI: "Punjabi",
    LANGID_RHAETO_ROMANIC: "Rhaeto Romanic",
    LANGID_ROMANIAN: "Romanian",
    LANGID_ROMANIAN_MOLDOVA: "Romanian Moldova",
    LANGID_RUSSIAN: "Russian",
    LANGID_RUSSIAN_MOLDOVA: "Russian Moldova",
    LANGID_SAMI_LAPPISH: "Sami Lappish",
    LANGID_SANSKRIT: "Sanskrit",
    LANGID_SERBIAN_CYRILLIC: "Serbian Cyrillic",
    LANGID_SERBIAN_LATIN: "Serbian Latin",
    LANGID_SESOTHO: "Sesotho",
    LANGID_SIMPLIFIED_CHINESE: "Simplified Chinese",
    LANGID_SINDHI: "Sindhi",
    LANGID_SINDHI_PAKISTAN: "Sindhi (Pakistan)",
    LANGID_SINHALESE: "Sinhalese",
    LANGID_SLOVAK: "Slovakian",
    LANGID_SLOVENIAN: "Slovenian",
    LANGID_SOMALI: "Somali",
    LANGID_SORBIAN: "Sorbian",
    LANGID_SPANISH: "Spanish",
    LANGID_SPANISH_ARGENTINA: "Spanish Argentina",
    LANGID_SPANISH_BOLIVIA: "Spanish Bolivian",
    LANGID_SPANISH_CHILE: "Spanish Chilean",
    LANGID_SPANISH_COLOMBIA: "Spanish Colombian",
    LANGID_SPANISH_COSTA_RICA: "Spanish Costa Rican",
    LANGID_SPANISH_DOMINICAN_REPUBLIC: "Spanish Dominican Republic",
    LANGID_SPANISH_ECUADOR: "Spanish Ecuadorian",
    LANGID_SPANISH_EL_SALVADOR: "Spanish El Salvadorian",
    LANGID_SPANISH_GUATEMALA: "Spanish Guatemala",
    LANGID_SPANISH_HONDURAS: "Spanish Honduran",
    LANGID_SPANISH_MODERN_SORT: "Spanish Modern Sort",
    LANGID_SPANISH_NICARAGUA: "Spanish Nicaraguan",
    LANGID_SPANISH_PANAMA: "Spanish Panamanian",
    LANGID_SPANISH_PARAGUAY: "Spanish Paraguayan",
    LANGID_SPANISH_PERU: "Spanish Peruvian",
    LANGID_SPANISH_PUERTO_RICO: "Spanish Puerto Rican",
    LANGID_SPANISH_URUGUAY: "Spanish Uruguayan",
    LANGID_SPANISH_VENEZUELA: "Spanish Venezuelan",
    LANGID_SUTU: "Sutu",
    LANGID_SWAHILI: "Swahili",
    LANGID_SWEDISH: "Swedish",
    LANGID_SWEDISH_FINLAND: "Swedish Finnish",
    LANGID_SWISS_FRENCH: "Swiss French",
    LANGID_SWISS_GERMAN: "Swiss German",
    LANGID_SWISS_ITALIAN: "Swiss Italian",
    LANGID_SYRIAC: "Syriac",
    LANGID_TAJIK: "Tajik",
    LANGID_TAMAZIGHT: "Tamazight",
    LANGID_TAMAZIGHT_LATIN: "Tamazight Latin",
    LANGID_TAMIL: "Tamil",
    LANGID_TATAR: "Tatar",
    LANGID_TELUGU: "Telugu",
    LANGID_THAI: "Thai",
    LANGID_TIBETAN: "Tibetan",
    LANGID_TIGRIGNA_ERITREA: "Tigrigna Eritrea",
    LANGID_TIGRIGNA_ETHIOPIC: "Tigrigna Ethiopic",
    LANGID_TRADITIONAL_CHINESE: "Traditional Chinese",
    LANGID_TSONGA: "Tsonga",
    LANGID_TSWANA: "Tswana",
    LANGID_TURKISH: "Turkish",
    LANGID_TURKMEN: "Turkmen",
    LANGID_UKRAINIAN: "Ukrainian",
    LANGID_URDU: "Urdu",
    LANGID_UZBEK_CYRILLIC: "Uzbek Cyrillic",
    LANGID_UZBEK_LATIN: "Uzbek Latin",
    LANGID_VENDA: "Venda",
    LANGID_VIETNAMESE: "Vietnamese",
    LANGID_WELSH: "Welsh",
    LANGID_XHOSA: "Xhosa",
    LANGID_YI: "Yi",
    LANGID_YIDDISH: "Yiddish",
    LANGID_YORUBA: "Yoruba",
    LANGID_ZULU: "Zulu"
}



import unittest
import struct
import pytest
import pytds
from pytds.smp import *
from utils import MockSock


smp_hdr = struct.Struct('<BBHLLL')


class SmpSessionsTests(unittest.TestCase):
    def setUp(self):
        self.sock = MockSock()
        self.mgr = SmpManager(self.sock)
        self.sess = self.mgr.create_session()
        self.buf = bytearray(b'0' * 100)
        self.sock.consume_output()

    def test_valid_data(self):
        self.sock.set_input([smp_hdr.pack(0x53, 8, 0, len(b'test') + 16, 1, 10) + b'test'])
        l = self.sess.recv_into(self.buf)
        assert self.buf[:l] == b'test'

    def test_invalid_flags(self):
        self.sock.set_input([smp_hdr.pack(0x53, 16, 0, 16, 1, 10)])
        with pytest.raises(pytds.Error) as excinfo:
            self.sess.recv_into(self.buf)
        assert 'Unexpected FLAGS' in str(excinfo.value)

    def test_syn_packet(self):
        """
        Server should not send SYN packets to a client, only client can send those
        """
        self.sock.set_input([smp_hdr.pack(0x53, 1, 0, 16, 1, 10)])
        with pytest.raises(pytds.Error) as excinfo:
            self.sess.recv_into(self.buf)
        assert 'Unexpected SYN' in str(excinfo.value)

    def test_data_after_fin(self):
        sess2 = self.mgr.create_session()
        self.sock.set_input([smp_hdr.pack(0x53, 4, 0, 16, 1, 10) + smp_hdr.pack(0x53, 8, 0, 16, 2, 10)])
        assert self.sess.recv_into(self.buf) == 0
        with pytest.raises(pytds.Error) as excinfo:
            sess2.recv_into(self.buf)
        assert 'Unexpected DATA packet from server' in str(excinfo.value)

    def test_fin_after_fin(self):
        sess2 = self.mgr.create_session()
        self.sock.set_input([smp_hdr.pack(0x53, 4, 0, 16, 1, 10) + smp_hdr.pack(0x53, 4, 0, 16, 2, 10)])
        assert self.sess.recv_into(self.buf) == 0
        with pytest.raises(pytds.Error) as excinfo:
            sess2.recv_into(self.buf)
        assert 'Unexpected FIN' in str(excinfo.value)

    def test_data_after_close(self):
        """should ignore data sent from server if we already send FIN packet"""
        self.sock.set_input([smp_hdr.pack(0x53, 8, 0, len(b'test') + 16, 1, 10) + b'test' +
                             smp_hdr.pack(0x53, 4, 0, 16, 1, 10)])
        assert self.sess.get_state() == SessionState.SESSION_ESTABLISHED
        self.sess.close()
        assert self.sess.get_state() == SessionState.CLOSED
        assert self.sess.recv_into(self.buf) == 0

    def test_close_twice(self):
        # this test is optional, maybe it does not behave like that
        self.sock.set_input([smp_hdr.pack(0x53, 4, 0, 16, 1, 10)])
        self.sess.close()
        self.sess.close()

    def test_ack_after_fin(self):
        sess2 = self.mgr.create_session()
        self.sock.set_input([smp_hdr.pack(0x53, 4, 0, 16, 1, 10) + smp_hdr.pack(0x53, 2, 0, 16, 2, 10)])
        assert self.sess.recv_into(self.buf) == 0
        with pytest.raises(pytds.Error) as excinfo:
            sess2.recv_into(self.buf)
        assert 'Unexpected ACK packet from server' in str(excinfo.value)

    def test_unexpected_eof(self):
        """
        Should raise EOF error if socket does not have enough data to fill SMP header
        """
        self.sock.set_input([b'0' * 10])
        with pytest.raises(pytds.Error) as excinfo:
            self.sess.recv_into(self.buf)
        assert 'Unexpected EOF' in str(excinfo.value)

    def test_invalid_id(self):
        self.sock.set_input([smp_hdr.pack(0, 4, 0, 16, 1, 10)])
        with pytest.raises(pytds.Error) as excinfo:
            self.sess.recv_into(self.buf)
        assert 'Invalid SMP packet signature' in str(excinfo.value)

    def test_invalid_session_id(self):
        self.sock.set_input([smp_hdr.pack(0x53, 0, 1, 0, 0, 0)])
        with pytest.raises(pytds.Error) as excinfo:
            self.sess.recv_into(self.buf)
        assert 'Invalid SMP packet session id' in str(excinfo.value)

    def test_invalid_wndw_value(self):
        self.sock.set_input([smp_hdr.pack(0x53, 0, 0, 0, 0, 0)])
        with pytest.raises(pytds.Error) as excinfo:
            self.sess.recv_into(self.buf)
        assert 'Invalid WNDW in packet from server' in str(excinfo.value)

    def test_invalid_seqnum_value(self):
        self.sock.set_input([smp_hdr.pack(0x53, 8, 0, 0, 500, 10)])
        with pytest.raises(pytds.Error) as excinfo:
            self.sess.recv_into(self.buf)
        assert 'Invalid SEQNUM in packet from server' in str(excinfo.value)

    def test_invalid_length(self):
        self.sock.set_input([smp_hdr.pack(0x53, 8, 0, 0, 1, 10)])
        with pytest.raises(pytds.Error) as excinfo:
            self.sess.recv_into(self.buf)
        assert 'Invalid LENGTH' in str(excinfo.value)

    def test_invalid_seqnum_in_data_packet(self):
        self.sock.set_input([smp_hdr.pack(0x53, 8, 0, 16, 0, 10)])
        with pytest.raises(pytds.Error) as excinfo:
            self.sess.recv_into(self.buf)
        assert 'Invalid SEQNUM' in str(excinfo.value)

    def test_invalid_seqnum_in_ack_packet(self):
        self.sock.set_input([smp_hdr.pack(0x53, 2, 0, 16, 1, 10)])
        with pytest.raises(pytds.Error) as excinfo:
            self.sess.recv_into(self.buf)
        assert 'Invalid SEQNUM' in str(excinfo.value)


def test_misc():
    sock = MockSock()
    mgr = SmpManager(sock)
    sess = mgr.create_session()
    repr(mgr)

    SessionState.to_str(SessionState.SESSION_ESTABLISHED)
    SessionState.to_str(SessionState.CLOSED)
    SessionState.to_str(SessionState.FIN_RECEIVED)
    SessionState.to_str(SessionState.FIN_SENT)

    mgr = SmpManager(sock, max_sessions=5)
    with pytest.raises(pytds.Error) as excinfo:
        for _ in range(10):
            mgr.create_session()
    assert "Can't create more MARS sessions" in str(excinfo.value)
